from __future__ import annotations

import json
from unittest.mock import patch

import pytest

from tatsu.util.asjson import asjson
from tatsu.util.fromjson import fromjson
from tatsu.ztyle import RGB, Color, Style, fmt, rgb
from tatsu.ztyle.xstyle import XStyle


@pytest.fixture(autouse=True)  # noqa: RUF076
def _tty_stdout():
    with patch("sys.stdout.isatty", return_value=True):
        yield


def test_default_empty():
    assert not str(Style())


def test_value_only():
    assert str(Style('hello')) == 'hello'


def test_fg_standard():
    for idx, sgr in enumerate(range(30, 38)):
        assert str(Style('x', fg=idx)) == f'\033[{sgr}mx\033[0m'


def test_fg_bright():
    for idx, sgr in enumerate(range(90, 98)):
        assert str(Style('x', fg=idx + 8)) == f'\033[{sgr}mx\033[0m'


def test_fg_256():
    assert str(Style('x', fg=16)) == '\033[38;5;16mx\033[0m'
    assert str(Style('x', fg=255)) == '\033[38;5;255mx\033[0m'


def test_bg_standard():
    for idx, sgr in enumerate(range(40, 48)):
        assert str(Style('x', bg=idx)) == f'\033[{sgr}mx\033[0m'


def test_bg_bright():
    for idx, sgr in enumerate(range(100, 108)):
        assert str(Style('x', bg=idx + 8)) == f'\033[{sgr}mx\033[0m'


def test_bg_256():
    assert str(Style('x', bg=16)) == '\033[48;5;16mx\033[0m'
    assert str(Style('x', bg=255)) == '\033[48;5;255mx\033[0m'


def test_value_with_fg():
    assert str(Style('hello', fg=1)) == '\033[31mhello\033[0m'


def test_modifier_bold():
    assert str(Style('x').bold()) == '\033[1mx\033[0m'


def test_modifier_dim():
    assert str(Style('x').dim()) == '\033[2mx\033[0m'


def test_modifier_italic():
    assert str(Style('x').italic()) == '\033[3mx\033[0m'


def test_modifier_underline():
    assert str(Style('x').underline()) == '\033[4mx\033[0m'


def test_modifier_blink():
    assert str(Style('x').blink()) == '\033[5mx\033[0m'


def test_modifier_inverse():
    assert str(Style('x').inverse()) == '\033[7mx\033[0m'


def test_modifier_hidden():
    assert str(Style('x').hidden()) == '\033[8mx\033[0m'


def test_modifier_strikethrough():
    assert str(Style('x').strikethrough()) == '\033[9mx\033[0m'


def test_modifier_combination():
    s = Style('x').bold().italic().underline()
    assert str(s) == '\033[1;3;4mx\033[0m'


def test_fg_with_modifier():
    s = Style('x').bold().fg(1)
    assert str(s) == '\033[1;31mx\033[0m'


def test_full_composition():
    s = Style('text', fg=2, bg=7, bold=True, italic=True)
    assert str(s) == '\033[1;3;32;47mtext\033[0m'


def test_named_color_method():
    s = XStyle('x').pink()
    assert str(s) == '\033[38;5;200mx\033[0m'


def test_named_color_bg_method():
    s = XStyle('x').pink_bg()
    assert str(s) == '\033[48;5;200mx\033[0m'


def test_chained_named_color():
    s = XStyle('x').bold().pink()
    assert str(s) == '\033[1;38;5;200mx\033[0m'


def test_repr_escapes():
    s = Style('hi', fg=1)
    r = repr(s)
    assert '\\e[31m' in r
    assert '\\e[0m' in r


def test_rgb_constructor():
    assert str(Style('x', fg=rgb(255, 0, 0))) == '\033[38;2;255;0;0mx\033[0m'


def test_rgb_constructor_bg():
    assert str(Style('x', bg=rgb(0, 255, 0))) == '\033[48;2;0;255;0mx\033[0m'


def test_rgb_both():
    s = Style('x', fg=rgb(255, 0, 0), bg=rgb(0, 0, 255))
    assert str(s) == '\033[38;2;255;0;0;48;2;0;0;255mx\033[0m'


def test_fg_rgb_method():
    s = Style('x').fg_rgb(128, 128, 128)
    assert str(s) == '\033[38;2;128;128;128mx\033[0m'


def test_bg_rgb_method():
    s = Style('x').bg_rgb(64, 64, 64)
    assert str(s) == '\033[48;2;64;64;64mx\033[0m'


def test_rgb_chained_with_modifier():
    s = Style('x').bold().fg_rgb(255, 0, 255)
    assert str(s) == '\033[1;38;2;255;0;255mx\033[0m'


def test_rgb_clamp():
    assert str(Style('x', fg=rgb(-10, 300, 128))) == '\033[38;2;0;255;128mx\033[0m'


def test_rgb_from_namedtuple():
    c = RGB(r=100, g=200, b=50)
    assert str(Style('x', fg=c)) == '\033[38;2;100;200;50mx\033[0m'


def test_default_no_escape_when_no_style():
    assert str(Style('plain')) == 'plain'


def test_no_color_suppresses_ansi(monkeypatch):
    monkeypatch.setenv("NO_COLOR", "1")
    s = Style('x', fg=1)
    assert str(s) == 'x'


def test_force_color_enables_ansi(monkeypatch):
    monkeypatch.setenv("FORCE_COLOR", "1")
    with patch("sys.stdout.isatty", return_value=False):
        s = Style('x', fg=1)
    assert str(s) == '\033[31mx\033[0m'


def test_force_enable_overrides_env(monkeypatch):
    monkeypatch.setenv("NO_COLOR", "1")
    s = Style('x', fg=1, color=Color(enable=True))
    assert str(s) == '\033[31mx\033[0m'


def test_no_color_empty_still_suppresses(monkeypatch):
    monkeypatch.setenv("NO_COLOR", "")
    s = Style('x', fg=1)
    assert str(s) == 'x'


def test_enabled_evaluates_dynamically(monkeypatch):
    _ = monkeypatch
    with patch("sys.stdout.isatty", return_value=True):
        s = Style('x', fg=1)
    assert str(s) == '\033[31mx\033[0m'
    with patch("sys.stdout.isatty", return_value=False):
        assert str(s) == 'x'


def test_no_color_dynamic(monkeypatch):
    s = Style('x', fg=1)
    assert str(s) == '\033[31mx\033[0m'
    monkeypatch.setenv("NO_COLOR", "1")
    assert str(s) == 'x'


def test_force_color_dynamic(monkeypatch):
    with patch("sys.stdout.isatty", return_value=False):
        s = Style('x', fg=1)
        assert str(s) == 'x'
    monkeypatch.setenv("FORCE_COLOR", "1")
    assert str(s) == '\033[31mx\033[0m'


def test_color_tty():
    with patch("sys.stdout.isatty", return_value=True):
        assert Color.tty().enabled is True
    with patch("sys.stdout.isatty", return_value=False):
        assert Color.tty().enabled is False


def test_color_always():
    assert Color.always().enabled is True


def test_color_never():
    assert Color.never().enabled is False


def test_color_is_terminal():
    with patch("sys.stdout.isatty", return_value=True):
        assert Color().is_terminal is True
    with patch("sys.stdout.isatty", return_value=False):
        assert Color().is_terminal is False


def test_color_supports_color():
    assert Color().supports_color is True


def test_color_supports_color_notty():
    with patch("sys.stdout.isatty", return_value=False):
        assert Color().supports_color is False


def test_color_supports_color_dumb_term(monkeypatch):
    monkeypatch.setenv("TERM", "dumb")
    assert Color().supports_color is False


def test_color_supports_color_emacs_term(monkeypatch):
    monkeypatch.setenv("TERM", "emacs")
    assert Color().supports_color is False


def test_color_style_factory():
    c = Color.always()
    s = c.style("hi", fg=1)
    assert str(s) == "\033[31mhi\033[0m"


def test_color_terminal_size():
    cols, lines = Color().terminal_size()
    assert isinstance(cols, int)
    assert isinstance(lines, int)
    assert cols > 0
    assert lines > 0


def test_fg_name_lookup():
    s = Style('x').fg_name('pink')
    assert str(s) == '\033[38;5;200mx\033[0m'


def test_bg_name_lookup():
    s = Style('x').bg_name('red')
    assert str(s) == '\033[48;5;196mx\033[0m'


def test_fg_css():
    s = Style('x').fg_css('crimson')
    assert str(s) == '\033[38;2;220;20;60mx\033[0m'


def test_bg_css():
    s = Style('x').bg_css('navy')
    assert str(s) == '\033[48;2;0;0;128mx\033[0m'


def test_fg_css_unknown():
    s = Style('x').fg_css('nonexistent')
    assert str(s) == 'x'


def test_fg_css_case_insensitive():
    s = Style('x').fg_css('DeepSkyBlue')
    assert str(s) == '\033[38;2;0;191;255mx\033[0m'


def test_color_style_factory_disabled():
    c = Color.never()
    s = c.style("hi", fg=1)
    assert str(s) == "hi"


def test_fmt_none():
    assert fmt("hello", "") == "hello"


def test_fmt_right_align():
    assert fmt("x", ">10") == "         x"


def test_fmt_left_align():
    assert fmt("x", "<10") == "x         "


def test_fmt_center():
    assert fmt("hi", "^10") == "    hi    "


def test_fmt_fill_char():
    assert fmt("x", "->10") == "---------x"


def test_fmt_truncate():
    assert fmt("hello world", ".5") == "hello"


def test_fmt_width_and_truncate():
    assert fmt("hello world", ">10.5") == "     hello"


def test_fmt_empty_text():
    assert fmt("", ">10") == "          "


def test_fmt_zero_width():
    assert fmt("x", "0") == "x"


def test_fmt_invalid_spec():
    assert fmt("x", "invalid") == "x"


def test_fmt_through_call():
    s = Style().fmt(">10")("x")
    assert str(s) == "         x"


def test_fmt_through_str():
    s = Style("x").fmt(">10")
    assert str(s) == "         x"


def test_fmt_with_ansi():
    s = Style(bold=True, color=Color.always()).fmt(">5")
    assert s.apply("x") == "\033[1m    x\033[0m"


def test_fmt_immutable():
    base = Style("x")
    styled = base.fmt(">10")
    assert base._fmt is None
    assert styled._fmt == ">10"


# def test_str_methods_return_plain_str():
#     """String methods on Style return plain str, not Style instances.

#     Since Style inherits from str, methods like .strip(), .upper(),
#     .replace() etc. operate at the str level and return str objects.
#     Chaining style methods after a str method will fail.
#     """
#     s = Style("  hello  ", bold=True)

#     stripped = s.strip()
#     assert stripped == "hello"
#     assert type(stripped) is str

#     upper = s.upper()
#     assert upper == "  HELLO  "
#     assert type(upper) is str

#     replaced = s.replace("hello", "world")
#     assert replaced == "  world  "
#     assert type(replaced) is str

#     with pytest.raises(AttributeError):
#         s.strip().bold()

#     s = s("hello")
#     assert str(s) == "\033[1mhello\033[0m"


def test_parse_plain():
    assert str(Style.from_raw("hello")) == "hello"


def test_parse_fg():
    s = Style.from_raw("\033[31mhello\033[0m")
    assert str(s) == "\033[31mhello\033[0m"


def test_parse_fg_bright():
    s = Style.from_raw("\033[91mhello\033[0m")
    assert str(s) == "\033[91mhello\033[0m"


def test_parse_bold_fg():
    s = Style.from_raw("\033[1;32mhello\033[0m")
    assert str(s) == "\033[1;32mhello\033[0m"


def test_parse_256():
    s = Style.from_raw("\033[38;5;200mhello\033[0m")
    assert str(s) == "\033[38;5;200mhello\033[0m"


def test_parse_rgb():
    s = Style.from_raw("\033[38;2;255;0;0mhello\033[0m")
    assert str(s) == "\033[38;2;255;0;0mhello\033[0m"


def test_parse_modifiers():
    s = Style.from_raw("\033[1;3;4mhello\033[0m")
    assert str(s) == "\033[1;3;4mhello\033[0m"


def test_parse_all_modifiers():
    s = Style.from_raw("\033[1;2;3;4;5;7;8;9mhello\033[0m")
    assert str(s) == "\033[1;2;3;4;5;7;8;9mhello\033[0m"


def test_parse_background():
    s = Style.from_raw("\033[41mhello\033[0m")
    assert str(s) == "\033[41mhello\033[0m"


def test_parse_background_256():
    s = Style.from_raw("\033[48;5;100mhello\033[0m")
    assert str(s) == "\033[48;5;100mhello\033[0m"


def test_parse_background_rgb():
    s = Style.from_raw("\033[48;2;0;0;128mhello\033[0m")
    assert str(s) == "\033[48;2;0;0;128mhello\033[0m"


def test_parse_backslash_e():
    s = Style.from_raw("\\e[31mhello\\e[0m")
    assert str(s) == "\033[31mhello\033[0m"


def test_fromjson_reconstructs_style():
    c = Color.always()
    original = Style("hello", fg=1, bold=True, color=c)
    payload = asjson(original)
    result = fromjson(payload)
    assert isinstance(result, Style)
    assert result is not original
    assert result._fg == original._fg
    assert result._bold == original._bold
    assert result._bg == original._bg


def test_full_json_roundtrip_reconstructs_style():
    c = Color.always()
    original = Style("hello", fg=1, bold=True, color=c)
    serialized = json.dumps(asjson(original))
    result = fromjson(json.loads(serialized))
    assert isinstance(result, Style)
    assert result._fg == original._fg
    assert result._bold == original._bold


def test_fromjson_parses_escape_strings():
    result = fromjson("\\e[31mhello\\e[0m")
    assert isinstance(result, Style)
    assert result._fg == 1


def test_repr_includes_fmt_text_after_colon():
    c = Color.always()
    s = Style("hello", fg=1, fmt=">10", color=c)
    r = repr(s)
    assert "f{hello:>10}" in r


def test_repr_no_fmt_no_extra_colon():
    c = Color.always()
    s = Style("hello", fg=1, color=c)
    r = repr(s)
    assert "hello" in r
    assert "f{" not in r


def test_fmt_roundtrip_with_styling():
    c = Color.always()
    original = Style("hello", fg=1, fmt=">10", color=c)
    payload = asjson(original)
    result = fromjson(payload)
    assert isinstance(result, Style)
    assert result._fmt == ">10"
    assert result.value == "hello"


def test_fmt_and_color_roundtrip():
    original = Style("hello", fg=2, bold=True, fmt=">10", color=Color.always())
    payload = asjson(original)
    result = fromjson(payload)
    assert isinstance(result, Style)
    assert result._fg == 2
    assert result._bold is True
    assert result._fmt == ">10"
    assert result._color._force_enable is None  # default color policy


def test_repr_fmt_roundtrips():
    c = Color.always()
    original = Style("hello", fmt=">10", color=c)
    r = repr(original)
    parsed = Style.from_raw(r)
    assert parsed._fmt == ">10"
    assert parsed.value == "hello"


def test_fromjson_reconstructs_nested_style():
    c = Color.always()
    original = {"title": Style("T", bold=True, color=c)}
    serialized = json.dumps(asjson(original))
    result = fromjson(json.loads(serialized))
    assert isinstance(result["title"], Style)
    assert result["title"] is not original["title"]
    assert result["title"]._bold is True



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .basic import *  # noqa: F403
from .rule import *  # noqa: F403
from .tatsumasu import *  # noqa: F403



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Protocol


class Payload(Protocol):
    path: Path
    payload: Any

    def raises(self) -> tuple[type[Exception], ...]:
        return ()


class StrPayload(str, Payload):
    __slots__ = ()

    @property
    def path(self) -> Path:  # type: ignore
        return Path(self)

    @property
    def payload(self) -> Any:  # type: ignore
        return Path(self).read_text()


@dataclass(slots=True)
class VisualPayload(Payload):
    path: Path
    payload: Any



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Any

from ..ctx import Ctx, Func
from ._base import ContextBase


class ChoiceContext(ContextBase):
    def __init__(self, ctx: Ctx):
        super().__init__(ctx)
        self.options: list[Func] = []
        self.result: Any = None

    def option(self, func: Func) -> Func:
        self.options.append(func)
        return func

    def parse(self, ctx: Ctx) -> Any:
        if not self.options:
            return None
        for opt in self.options:
            with ctx.option():
                ctx.expcall(opt)
        raise self.expectedexcept(ctx)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import sys

import pytest

import tatsu
from tatsu.api import compile, parse
from tatsu.exceptions import FailedParse, FailedToken
from tatsu.objectmodel import Node, synth
from tatsu.objectmodel.builder import (
    BuilderConfig,
    ModelBuilderSemantics,
    TypeResolutionError,
)


class MyNode:
    def __init__(self, _ast):
        pass


class BType:
    pass


class AType(BType):
    pass


def test_semantics_not_class():
    grammar = r"""
        start::sum = {number}+ $ ;
        number::int = /\d+/ ;
    """
    text = '5 4 3 2 1'
    bad_semantics = ModelBuilderSemantics  # NOTE: the class
    semantics = ModelBuilderSemantics()  # NOTE: the class

    with pytest.raises(
        TypeError,
        match=r'semantics must be an object instance or None.*',
    ):
        compile(grammar, semantics=bad_semantics)
        compile(grammar, semantics=bad_semantics)

    model = compile(grammar, 'test')
    with pytest.raises(
        TypeError,
        match=r'semantics must be an object instance or None.*',
    ):
        model.parse(text, semantics=bad_semantics)

    ast = model.parse(text, semantics=semantics)
    assert ast == 15


def test_builder_semantics():
    grammar = r"""
        start::sum = {number}+ $ ;
        number::int = /\d+/ ;
    """
    text = '5 4 3 2 1'

    semantics = ModelBuilderSemantics()
    model = compile(grammar, 'test')
    ast = model.parse(text, semantics=semantics)
    assert ast == 15

    import functools

    dotted = functools.partial(str.join, '.')
    # WARNING: this defeats all TatSu knows about functions and types
    #  dotted.__name__ = 'dotted'

    grammar = r"""
        start::dotted = {number}+ $ ;
        number = /\d+/ ;
    """

    semantics = ModelBuilderSemantics(constructors=[dotted])
    model = compile(grammar, 'test')
    with pytest.raises(
        TypeResolutionError,
        match=r"Could not find constructor for type 'dotted'",
    ):
        ast = model.parse(text, semantics=semantics)
        assert ast == '5.4.3.2.1'


def test_builder_subclassing():
    registry = getattr(synth, '__registry')

    grammar = """
        @@grammar :: Test
        start::A::B::C = $ ;
    """

    model = tatsu.compile_to_parser(grammar)
    model.parse('')

    print(f'{registry=}')
    A = registry['A']
    B = registry['B']
    C = registry['C']

    assert issubclass(A, B) and issubclass(A, synth.SynthNode) and issubclass(A, Node)
    assert issubclass(B, C) and issubclass(B, synth.SynthNode) and issubclass(A, Node)
    assert issubclass(C, synth.SynthNode) and issubclass(C, Node)


def test_builder_basetype_codegen():
    grammar = """
        @@grammar :: Test
        start::A::B::C = a:() b:() $ ;
        second::D::A = ();
        third = ();
    """

    from tatsu.api import to_python_model

    src = to_python_model(grammar, basetype=MyNode)
    # print(src[:1000])

    globals = {}
    exec(src, globals)  # pylint: disable=W0122
    semantics = globals['TestModelBuilderSemantics']()

    A = globals['A']
    B = globals['B']
    C = globals['C']
    D = globals['D']

    model = compile(grammar, semantics=semantics)
    ast = model.parse('', semantics=semantics)
    # print(f'AST({type(ast)}=', ast)

    assert isinstance(ast, MyNode), A.__mro__
    assert isinstance(ast, (A, B, C))
    assert hasattr(ast, 'a')
    assert hasattr(ast, 'b')

    assert issubclass(D, A | B | C)


def test_optional_attributes():
    grammar = r"""
        foo::Foo = left:identifier [ ':' right:identifier ] $ ;
        identifier = /\w+/ ;
    """

    grammar = compile(grammar)

    a = grammar.parse('foo : bar', semantics=ModelBuilderSemantics())
    assert a.left == 'foo'
    assert a.right == 'bar'

    b = grammar.parse('foo', semantics=ModelBuilderSemantics())
    assert b.left == 'foo'
    assert b.right is None


def test_constant_math():
    grammar = r"""
        start = a:`7` b:`2` @:```{a} / {b}``` $ ;
    """
    result = parse(grammar, '', trace=True)
    assert result <= 3.5 <= result


def test_constant_deep_eval():
    grammar = r"""
        start =
            a:A b:B
            @:```{a} / {b}```
            $
        ;

        A = number @:`7` ;
        B = number @:`0` ;
        number::int = /\d+/ ;
    """
    model = compile(grammar)

    with pytest.raises(
        FailedParse,
        match=r'Error evaluating constant.*ZeroDivisionError',
    ):
        # NOTE: only with multiple evaluation passes on constants
        model.parse('42 84', trace=True)


def test_builder_types():
    grammar = """
        @@grammar :: Test
        start::AType::BType = $ ;
    """

    builderconfig = BuilderConfig(basetype=BType, constructors=[AType, BType])
    ast = parse(grammar, '', builderconfig=builderconfig)
    assert type(ast) is AType
    assert isinstance(ast, BType)
    assert not isinstance(ast, synth.SynthNode)


def test_builder_nodedefs():
    grammar = """
        @@grammar :: Test
        start::AType::BType = $ ;
    """

    thismodule = sys.modules[__name__]
    builderconfig = BuilderConfig(typedefs=[thismodule], synthok=False)
    ast = parse(grammar, '', builderconfig=builderconfig)
    assert type(ast) is AType
    assert isinstance(ast, BType)
    assert not isinstance(ast, synth.SynthNode)


def test_ast_per_option():
    grammar = """
        start = options $ ;

        options =
            | a:'a' [b:'b']
            | c:'c' [d:'d']
            ;
    """

    # NOTE:
    #   Prove each option in a choide has its own version of the AST

    ast = parse(grammar, 'a b')
    assert ast == {'a': 'a', 'b': 'b'}
    assert 'c' not in ast
    assert 'd' not in ast

    ast = parse(grammar, 'c d')
    assert ast == {'c': 'c', 'd': 'd'}
    assert 'a' not in ast
    assert 'b' not in ast

    ast = parse(grammar, 'a')
    assert ast == {'a': 'a', 'b': None}
    assert 'b' in ast
    assert 'c' not in ast
    assert 'd' not in ast

    ast = parse(grammar, 'c')
    assert ast == {'c': 'c', 'd': None}
    assert 'd' in ast
    assert 'a' not in ast
    assert 'b' not in ast


def test_ast_names_accumulate():
    grammar = """
        start: options $

        options:
            | a='a' ([b='b'] {x='x'})
            | c='c' ([d='d'] y={'y'})
            | w={} f='f'
            | g=() g='g'
            | h={} h='h'
            | i+='i'
            | [j+='j'] k='k'
    """

    # NOTE:
    #   Prove named elements accumulat
    ast = parse(grammar, 'a')
    assert ast == {'a': 'a', 'b': None, 'x': None}

    ast = parse(grammar, 'a x')
    assert ast == {'a': 'a', 'b': None, 'x': 'x'}

    ast = parse(grammar, 'a x x')
    assert ast == {'a': 'a', 'b': None, 'x': ['x', 'x']}

    # NOTE:
    #   Prove naming closures always closure
    ast = parse(grammar, 'c')
    assert ast == {'c': 'c', 'd': None, 'y': []}

    ast = parse(grammar, 'c y')
    assert ast == {'c': 'c', 'd': None, 'y': ['y']}

    ast = parse(grammar, 'c y y')
    assert ast == {'c': 'c', 'd': None, 'y': ['y', 'y']}

    ast = parse(grammar, 'f')
    assert ast == {'f': 'f', 'w': []}

    ast = parse(grammar, 'g')
    assert ast == {'g': [(), 'g']}

    ast = parse(grammar, 'h')
    assert ast == {'h': [[], 'h']}

    ast = parse(grammar, 'i')
    assert ast == {'i': ['i']}

    ast = parse(grammar, 'k')
    assert ast == {'j': [], 'k': 'k'}


def test_cut_scope():
    grammar = """
        start = (failcut | failchoice | succeed) $ ;

        failcut = 'a' ~ 'y' ;

        failchoice =
            | 'a' ~ 'b'
            | 'a' 'c' 'd'
            ;

        succeed = ('a' ~ 'y' | 'b' 'z') | 'a' 'x' ;
    """

    ast = parse(grammar, 'a x')
    assert ast == ['a', 'x']

    ast = parse(grammar, 'a b')
    assert ast == ['a', 'b']

    ast = parse(grammar, 'a y')
    assert ast == ['a', 'y']

    with pytest.raises(FailedToken, match=r"expecting 'b'"):
        ast = parse(grammar, 'a c d')
        assert ast == ['a', 'c', 'd']


if __name__ == '__main__':
    test_cut_scope()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .. import peg as g
from .._version import __version__ as version
from ..util.common import typename
from ..util.indent import IndentPrintMixin
from .boilerplt import FOOTER
from .ngparser_gen import keywordsgen, textinputgen


def parsermodel_gen(model: g.Grammar, name: str | None = None) -> str:
    model = model.optimized()
    generator = ParseWithModelGenerator()
    return generator.generate_parser(model, name=name)


def PARSER(name: str) -> str:
    return f"""\
    #!/usr/bin/env python3
    #
    #  CAVEAT UTILITOR
    #
    #  This file was automatically generated by 竜TatSu v{version}
    #
    #     https://pypi.python.org/pypi/tatsu/
    #
    #  Any changes you make to it will be overwritten the next time
    #  the file is generated.
    #
    # noqa # type: ignore # ruff: noqa
    # copyright: ignore
    # fmt: off

    from __future__ import annotations
    from typing import Any

    from tatsu.config import ParserConfig
    from tatsu.contexts import CanParse
    from tatsu.input import Text
    from tatsu.input.buffer import Buffer
    from tatsu.input.textlines import TextLines
    from tatsu.peg import *


    class {name}Parser(CanParse):
        def __init__(self, config: Any = None, **settings: Any):
            self.config = ParserConfig.new(config, **settings)
            self.model = GRAMMAR_MODEL

        def parse(
            self,
            text: str | Text,
            /,
            *,
            start: str | None = None,
            asmodel: bool = True,
            config: Any = None,
            **settings: Any,
        ) -> Any:
            # NOTE
            #   Copy the grammar so the configuration is unique to this parse
            #   and one parse doesn't leak settings to another.
            #   There may be also configurations that are unique to this parse.
            model = Grammar(
                name=self.model.name,
                rules=self.model.rules,
                directives=self.model.directives,
                keywords=self.model.keywords,
                config=self.config,
            )
            config = ParserConfig.new(config, **settings)
            return model.parse(text, asmodel=asmodel, config=config)
    """


class ParseWithModelGenerator(IndentPrintMixin):
    def generate_parser(self, grammar: g.Grammar, name: str | None = None) -> str:
        name = name or grammar.name or ''
        self.clear()
        self.print()

        self.print(PARSER(name))
        self.print()
        self.print()

        self.print(f'GRAMMAR_MODEL: {typename(grammar)} = (')
        with self.indent():
            self.print(repr(grammar.optimized()))
        self.print(')')
        self.print()
        self.print()

        self.print(textinputgen(grammar, name))
        self.print()

        self.print(keywordsgen(grammar))
        self.print()

        self.print(FOOTER(name))

        return f"{self.printed_text().strip()}\n"



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from argparse import ArgumentParser


def add_grammar_options(parser: ArgumentParser):
    parser.add_argument(
        "-z",
        "--optimized",
        action="store_true",
        default=False,
        help="Use the optimized version of the grammar",
    )
    parser.add_argument(
        "-m",
        "--name",
        default="",
        dest="name",
        help="Previx for the name of geneated parsers",
    )
    format = parser.add_mutually_exclusive_group()
    format.add_argument(
        "-j",
        "--json",
        action="store_true",
        dest="json",
        default=True,
        help="output the grammar model in JSON format (default)",
    )
    format.add_argument(
        "-p",
        "--pretty",
        action="store_true",
        dest="pretty",
        help="output the grammar in pretty-printed EBNF format",
    )
    format.add_argument(
        '-y',
        '--pretty-lean',
        help='like --pretty, but without name= or [Parameter] annotations',
        action='store_true',
    )
    format.add_argument(
        "-r",
        "--railroads",
        action="store_true",
        dest="railroads",
        help="output a railroad diagram of the grammar",
    )
    format.add_argument(
        '-g',
        '--object-model',
        help='an AST model from the class names as rule parameters',
        action='store_true',
    )
    format.add_argument(
        "-e",
        "--generate-parser",
        action="store_true",
        help='generate a procedural parser from the grammar',
    )
    format.add_argument(
        "-x",
        "--parser-model",
        action="store_true",
        help='generate a model-based parser from the grammar',
    )



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from itertools import starmap
from typing import NamedTuple


class LineInfo(NamedTuple):
    source: str
    line: int
    col: int
    start: int
    end: int
    text: str


class LineIndexInfo(NamedTuple):
    filename: str
    line: int

    @staticmethod
    def block_index(name, n) -> list[LineIndexInfo]:
        return list(starmap(LineIndexInfo, zip(n * [name], range(n), strict=False)))


class PosLine(NamedTuple):
    startpos: int
    lineno: int
    length: int

    @staticmethod
    def build_line_cache(lines, size):
        # an index from original positions to PosLine entries
        if not lines:
            return [], 1

        cache = []
        n = 0
        i = 0
        for n, line in enumerate(lines):
            pl = PosLine(i, n, len(line))
            for _ in line:
                cache.append(pl)  # noqa: PERF401
            i += len(line)

        n += 1
        if lines[-1][-1] in {'\r', '\n'}:
            n += 1
        cache.append(PosLine(i, n, 0))

        # the range depends on line[-1] ending in a newline
        endrange = range(len(lines), 2 + len(lines))
        assert n in endrange
        assert len(cache) == 1 + size

        return cache, n



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from ..packetz import PacketLike
from .legacy import parallel_proc, processing_loop
from .parproc import (
    Progress,
    parproc,
)
from .payload import Payload, StrPayload, VisualPayload
from .pmap import GIL_DISABLED, HAS_MULTITHREADING_SUPPORT
from .result import Result
from .summary import show_summary
from .visual import parproc_visual


__all__ = [
    'GIL_DISABLED',
    'HAS_MULTITHREADING_SUPPORT',
    'PacketLike',
    'Payload',
    'Progress',
    'Result',
    'StrPayload',
    'VisualPayload',
    'parallel_proc',
    'parproc',
    'parproc_visual',
    'processing_loop',
    'show_summary',
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import math

import pytest

import tatsu
from tatsu.exceptions import FailedParse


def test_name_matches_identifier():
    ast = tatsu.parse('start = @name $ ;', 'hello')
    assert ast == 'hello'


def test_name_matches_underscore_start():
    ast = tatsu.parse('start = @name $ ;', '_hello')
    assert ast == '_hello'


def test_name_rejects_digit_start():
    with pytest.raises(FailedParse):
        tatsu.parse('start = @name $ ;', '1hello')


def test_int_matches_signed():
    ast = tatsu.parse('start = @int $ ;', '+42')
    assert ast == 42
    assert isinstance(ast, int)


def test_int_matches_negative():
    ast = tatsu.parse('start = @int $ ;', '-7')
    assert ast == -7


def test_int_matches_unsigned():
    ast = tatsu.parse('start = @int $ ;', '99')
    assert ast == 99


def test_int_rejects_float():
    with pytest.raises(FailedParse):
        tatsu.parse('start = @int $ ;', '3.14')


def test_int_rejects_empty():
    with pytest.raises(FailedParse):
        tatsu.parse('start = @int $ ;', '')


def test_uint_matches_unsigned():
    ast = tatsu.parse('start = @uint $ ;', '42')
    assert ast == 42
    assert isinstance(ast, int)


def test_uint_rejects_negative():
    with pytest.raises(FailedParse):
        tatsu.parse('start = @uint $ ;', '-5')


def test_uint_rejects_plus():
    with pytest.raises(FailedParse):
        tatsu.parse('start = @uint $ ;', '+5')


def test_uint_with_underscores():
    ast = tatsu.parse('start = @uint $ ;', '1_000_000')
    assert ast == 1000000


def test_float_matches_simple():
    ast = tatsu.parse('start = @float $ ;', str(math.pi))
    assert math.isclose(ast, math.pi)
    assert isinstance(ast, float)


def test_float_matches_negative():
    ast = tatsu.parse('start = @float $ ;', '-2.5')
    assert math.isclose(ast, -2.5)


def test_float_rejects_text():
    with pytest.raises(FailedParse):
        tatsu.parse('start = @float $ ;', 'abc')


def test_bool_matches_true():
    ast = tatsu.parse('start = @bool $ ;', 'true')
    assert ast is True


def test_bool_matches_capitalized_true():
    ast = tatsu.parse('start = @bool $ ;', 'True')
    assert ast is True


def test_meta_in_named():
    ast = tatsu.parse('start = value=@int $ ;', '42')
    assert ast == {'value': 42}


def test_meta_in_sequence():
    ast = tatsu.parse('start = @name "=" @int $ ;', 'x=7')
    assert ast == ['x', '=', 7]


def test_meta_as_atom():
    ast = tatsu.parse('start = @name $ ;', 'hello')
    assert ast == 'hello'



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from ..ctx import Ctx, Func
from .exp import ExpContext


class ExpWithSepContext(ExpContext):
    def __init__(self, ctx: Ctx):
        super().__init__(ctx)
        self._sep: Func | None = None

    @property
    def sep_func(self) -> Func:
        if self._sep is None:
            raise RuntimeError('.sep not set')
        return self._sep

    def sep(self, func: Func) -> Func:
        if self._sep is not None:
            raise RuntimeError('.sep already set')
        self._sep = func
        return func



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re
import sys
from functools import cache
from typing import Any


if sys.version_info >= (3, 13):
    from re import PatternError
else:
    PatternError = re.error


@cache
def cached_re_compile(
    pattern: str | bytes | re.Pattern,
    /,
    flags: int = 0,
) -> re.Pattern:
    if isinstance(pattern, re.Pattern):
        return pattern
    pattern = str(pattern)
    return re.compile(pattern, flags=flags)


def regexpp(regex: Any) -> str:
    """
    Returns a printable version of the regexp pattern as a Python raw string.
    Validates input and ensures generated output is syntactically valid.
    """
    # by Gemini (2026-02-04 - 2026-02-07)
    # by [apalala@gmail.com](https://github.com/apalala)

    pattern_text = regex.pattern if hasattr(regex, "pattern") else str(regex)

    try:
        re.compile(pattern_text)
    except PatternError as e:
        raise ValueError(
            f"Invalid regex passed to regexp(): {pattern_text!r}\n{e}",
        ) from e

    ctrl_map: dict[str, str] = {
        "\n": r"\n",
        "\r": r"\r",
        "\t": r"\t",
        "\v": r"\v",
        "\f": r"\f",
        "\b": r"\b",
        "\a": r"\a",
        "\0": r"\0",
    }

    result = "".join(ctrl_map.get(c, c) for c in pattern_text)

    # Handle trailing backslashes (odd count check for raw string safety)
    if result.endswith("\\") and (len(result) - len(result.rstrip("\\"))) % 2 != 0:
        result += "\\"

    if result.endswith("'") or result.count("'") > result.count('"'):
        output = f'r"{re.sub(r'(?<!\\)"', r"\"", result)}"'
    else:
        output = f"r'{re.sub(r"(?<!\\)'", r"\'", result)}'"

    try:
        evaluated = eval(output)  # noqa: S307
        re.compile(evaluated)
    except SyntaxError as e:
        raise RuntimeError(
            f"regexp() generated invalid Python syntax: {output}\n{e}",
        ) from e
    except PatternError as e:
        raise RuntimeError(
            f"regexp() generated an invalid regex pattern: {output}\n{e}",
        ) from e
    except Exception as e:
        raise RuntimeError(f"Unexpected error evaluating output: {output}\n{e}") from e

    return output



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations  # noqa: I001

import re
from typing import Any

from ..config import ParserConfig
from ..peg import GrammarSemantics
from ..semantics import ASTSemantics

# from .bootparser import TatSuBootstrapParser
from .bootstrap import (
    TatSuBootstrapBuffer,
    TatSuBootstrapText,
    TatSuBootstrapParser,
)


PRAGMA_RE = r'^\s*#include.*$'


class TatSuText(TatSuBootstrapText): ...


class TatSuBuffer(TatSuBootstrapBuffer):
    def __init__(
        self,
        text: str,
        /,
        filename: str | None = None,
        config: ParserConfig | None = None,
        **settings: Any,
    ):
        filename = filename or settings.pop('source', None)
        config = ParserConfig.new(config=config, source=filename, **settings)
        super().__init__(text, config=config)

    def process_block(self, name, lines, index, **_kwargs):
        i = 0
        while i < len(lines):
            line = lines[i]
            if re.match(PRAGMA_RE, line):
                directive, arg = line.split('#', 1)[1], ''
                if '::' in directive:
                    directive, arg = directive.split('::', 1)
                directive, arg = directive.strip(), arg.strip()
                i = self.pragma(name, directive, arg, lines, index, i)
            else:
                i += 1
        return lines, index

    def pragma(self, source, name, arg, lines, index, i):
        # we only recognize the 'include' pragama
        if name == 'include':
            filename = arg.strip('\'"')
            return self.include_file(source, filename, lines, index, i, i + 1)
        else:
            return i + 1  # will be treated as a directive by the parser


class TatSuParser(TatSuBootstrapParser):
    def __init__(
        self,
        name: str | None = None,
        config: ParserConfig | None = None,
        semantics=None,
        **settings: Any,
    ):
        if semantics is None:
            semantics = ASTSemantics()
        config = ParserConfig.new(
            config=config,
            name=name,
            semantics=semantics,
            **settings,
        )
        super().__init__(config=config)


class TatSuParserGenerator(TatSuBootstrapParser):
    def __init__(
        self,
        name: str | None = None,
        semantics: Any = None,
        **settings: Any,
    ):
        if isinstance(semantics, type):
            raise TypeError(
                f'semantics must be an object instance or None, not class {semantics!r}',
            )
        if not semantics:
            semantics = GrammarSemantics(name=name)
        config = ParserConfig.new(
            name=name,
            semantics=semantics,
            **settings,
        )
        super().__init__(config=config)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from pathlib import Path
from typing import Any

from ..api import compile
from ..peg import Grammar
from .cfg import CLIError


type Results = list[tuple[str, Any]]


def load_grammar(path: str | Path) -> Grammar:
    """Load a Grammar from an .ebnf or .json file."""

    p = Path(path)
    try:
        source = p.read_text(encoding="utf-8")
    except FileNotFoundError as e:
        raise CLIError(str(e)) from e
    if p.suffix == ".json":
        return Grammar.loads(source)

    return compile(source)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import dataclasses
import time
from collections.abc import Generator
from contextlib import contextmanager
from datetime import UTC, datetime
from pathlib import Path


__all__ = [
    'iso_logpath',
    'iso_timestamp',
    'timer',
    'Timing',
]


def iso_timestamp() -> str:
    """
    Returns an ISO 8601 string with UTC timezone (Z).
    Example: '2026-02-15T22-15-01Z'
    """
    now = datetime.now(UTC)
    isostr = now.isoformat(timespec='seconds')
    return isostr.replace(':', '-').replace('+00-00', 'Z')


def iso_logpath(prefix: str, basdir: str = "log", suffix='.log') -> Path:
    """Helper to generate a pathlib.Path using the ISO timestamp."""
    ts = iso_timestamp()
    logdir = Path(basdir)
    logdir.mkdir(parents=True, exist_ok=True)
    suffix = suffix if suffix.startswith('.') else '.' + suffix
    return logdir / f"{prefix}_{ts}{suffix}"


@dataclasses.dataclass(slots=True)
class Timing:
    start: int = 0
    _last_lap: int = 0
    _end: int = -1

    def __post_init__(self):
        now = time.monotonic_ns()
        self.start = now
        self._last_lap = now

    @property
    def delta(self) -> float:
        if self._end >= 0:
            ns_delta = self._end - self.start
        else:
            ns_delta = time.monotonic_ns() - self.start
        return ns_delta / 1_000_000_000

    def lap(self) -> float:
        if self._end >= 0:
            return 0.0
        now = time.monotonic_ns()
        ns_duration = now - self._last_lap
        self._last_lap = now
        return ns_duration / 1_000_000_000

    def __str__(self) -> str:
        d = self.delta
        return (
            f"{int(d // 60)}m {d % 60:.2f}s"
            if d >= 60
            else f"{d:.3f}s"
            if d >= 1
            else f"{d * 1000:.2f}ms"
        )


@contextmanager
def timer() -> Generator[Timing, None, None]:
    res = Timing()
    try:
        yield res
    finally:
        res._end = time.monotonic_ns()



import sys
import threading


# Global registry to ensure one SyncStream per physical stream object
_SYNC_STREAMS = {}
_REGISTRY_LOCK = threading.Lock()


class SyncStream:
    """Wraps a stream to ensure atomic write and flush operations."""

    def __init__(self, stream):
        self._stream = stream
        self._lock = threading.Lock()

    def write(self, data):
        with self._lock:
            return self._stream.write(data)

    def flush(self):
        with self._lock:
            return self._stream.flush()

    def __getattr__(self, name):
        return getattr(self._stream, name)


def get_sync_stream(stream):
    """Retrieves or creates a SyncStream for the given stream."""
    if isinstance(stream, SyncStream):
        return stream

    stream_id = id(stream)
    with _REGISTRY_LOCK:
        if stream_id not in _SYNC_STREAMS:
            _SYNC_STREAMS[stream_id] = SyncStream(stream)
        return _SYNC_STREAMS[stream_id]


class Logger:
    def __init__(self, fmt="[{name}] {level}: {message}", stream=sys.stderr):
        self.fmt = fmt
        self.stream = get_sync_stream(stream)
        self.defaults = {"name": "tatsu"}

    def log(self, level, message, **kwargs):
        context = {**self.defaults, "level": level, "message": message, **kwargs}
        # Format outside the lock, write inside the lock
        self.stream.write(self.fmt.format(**context) + "\n")
        self.stream.flush()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""
Process-safe packet queue for cross-worker communication.

Defines a Packet protocol, a default dataclass implementation, and
sync/async receivers for draining a multiprocessing.Queue.
"""

from __future__ import annotations

from .packet import Packet, PacketLike, WithID
from .queue import PacketzQueue


__all__ = [
    'Packet',
    'PacketLike',
    'PacketzQueue',
    'WithID',
]



# Copyright (c) 2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: MIT OR Apache-2.0

"""
json - Direct Value to Grammar translator

This module translates json directly to Grammar,
bypassing the TatSuModel deserializer which fails on modified JSON.
"""

from __future__ import annotations

import json
from typing import Any

from .. import grammars as g
from ..exceptions import TatSuException
from ..util import typename
from ..util.json import ensure_dict


class JsonError(TatSuException):
    pass


class JsonSerializationHelper:
    def __init__(self, value: Any, path: list[str] | None = None):
        self.value = value
        self.path = path if path is not None else []

    def _push(self, class_name: str) -> JsonSerializationHelper:
        new_path = [*self.path, class_name]
        return JsonSerializationHelper(self.value, new_path)

    def _error(self, msg: str) -> JsonError:
        path_str = " -> ".join(self.path)
        if not path_str:
            return JsonError(msg)
        return JsonError(f"{msg} at {path_str}")

    def _get_obj(self) -> dict[str, Any]:
        if isinstance(self.value, dict):
            return self.value
        raise self._error("Expected object")

    def _get_class(self) -> str:
        obj = self._get_obj()
        class_val = obj.get("__class__")
        if class_val is None:
            raise self._error("Missing __class__")
        if isinstance(class_val, str):
            return class_val
        raise self._error("Expected string for __class__")

    def _get_string(self, field: str) -> str:
        obj = self._get_obj()
        val = obj.get(field)
        if val is None:
            raise self._error(f"Missing field: {field}")
        if isinstance(val, str):
            return val
        raise self._error(f"Expected string for field: {field}")

    def _opt_str(self, field: str) -> str | None:
        obj = self._get_obj()
        val = obj.get(field)
        if val is None:
            return None
        if isinstance(val, str):
            return val
        return None

    def _opt_bool(self, field: str, default: bool) -> bool:
        obj = self._get_obj()
        val = obj.get(field)
        if val is None:
            return default
        if isinstance(val, bool):
            return val
        return default

    def _opt_int(self, field: str) -> int | None:
        obj = self._get_obj()
        val = obj.get(field)
        if val is None:
            return None
        if isinstance(val, (int, float)):
            return int(val)
        return None

    def _get_array(self, field: str) -> list[JsonSerializationHelper]:
        obj = self._get_obj()
        arr_val = obj.get(field)
        if arr_val is None:
            raise self._error(f"Missing or not array: {field}")
        if not isinstance(arr_val, list):
            raise self._error(f"Expected array for field: {field}")

        result = []
        for i, v in enumerate(arr_val):
            if isinstance(v, dict) and "__class__" in v:
                class_name = v["__class__"]  # ty: ignore
                label = f"{field}[{i}]:{class_name}"
            else:
                label = f"{field}[{i}]"
            result.append(self._push(label)._with_value(v))
        return result

    def _get_nested(self, field: str) -> JsonSerializationHelper:
        obj = self._get_obj()
        if field not in obj:
            raise self._error(f"Missing field: {field}")
        value = obj[field]

        if isinstance(value, dict) and "__class__" in value:
            class_name = value["__class__"]
            nested_path = f"{field}:{class_name}"
        else:
            nested_path = field

        return self._push(nested_path)._with_value(value)

    def _with_value(self, value: Any) -> JsonSerializationHelper:
        return JsonSerializationHelper(value, self.path)


def _parse_directives(directives: dict[str, Any] | None) -> dict[str, Any]:
    if directives is None:
        return {}

    result = {}
    for key, val in directives.items():
        if isinstance(val, str):
            result[key] = val
        elif isinstance(val, bool):
            result[key] = str(val)
        elif isinstance(val, (int, float)):
            result[key] = str(val)
        else:
            result[key] = str(val)
    return result


def loads_grammar(json_str: str) -> g.Grammar:
    """Parse JSON string and return a Grammar object."""
    value = json.loads(json_str)
    return load_grammar(value)


def load_grammar(value: Any) -> g.Grammar:
    """Parse JSON value and return a Grammar object."""
    value = ensure_dict(value)
    path = JsonSerializationHelper(value)
    class_name = path._get_class()

    if class_name != typename(g.Grammar):
        raise path._error(f"Expected {typename(g.Grammar)} root")

    name = path._get_string("name")

    rules_array = path._get_array("rules")
    rule_list = []
    for i, rule_path in enumerate(rules_array):
        try:
            rule = rule_from_json_path(rule_path)
            rule_list.append(rule)
        except JsonError as e:
            raise JsonError(f"rules[{i}]: {e}") from e

    directives_obj = path._get_obj().get("directives")
    if directives_obj is None:
        directives = {}
    elif isinstance(directives_obj, dict):
        directives = _parse_directives(directives_obj)
    else:
        directives = {}

    keywords: list[str] = []
    if isinstance(path._get_obj().get("keywords"), list):
        keywords = [str(k) for k in path._get_obj()["keywords"]]

    grammar = g.Grammar(rules=rule_list, name=name, directives=directives)
    if keywords:
        grammar.configure(keywords=keywords)
    grammar.initialize()
    return grammar


def rule_from_json_value(value: dict[str, Any]) -> g.Rule:
    """Parse JSON value (dict) and return a Rule object."""
    path = JsonSerializationHelper(value)
    return rule_from_json_path(path)


def rule_from_json_path(path: JsonSerializationHelper) -> g.Rule:
    """Parse JSON path and return a Rule object."""
    class_name = path._get_class()

    if class_name != typename(g.Rule):
        raise path._error(f"Expected {typename(g.Rule)}")

    name = path._get_string("name")
    exp_path = path._get_nested("exp")
    rhs = exp_from_json_path(exp_path)

    params: list[str] = []
    if isinstance(path._get_obj().get("params"), list):
        params = [str(p) for p in path._get_obj()["params"]]

    no_memo = path._opt_bool("no_memo", False)
    is_name = path._opt_bool("is_name", False)
    is_tokn = path._opt_bool("is_tokn", False)
    is_memo = path._opt_bool("is_memo", True)
    is_lrec = path._opt_bool("is_lrec", False)

    return g.Rule(
        name=name,
        exp=rhs,
        params=params,
        is_name=is_name,
        is_tokn=is_tokn,
        is_memo=is_memo,
        is_lrec=is_lrec,
        no_memo=no_memo,
    )


def exp_from_json_value(value: dict[str, Any]) -> g.Model:
    """Parse JSON value (dict) and return an expression Model."""
    path = JsonSerializationHelper(value)
    return exp_from_json_path(path)


def exp_from_json_path(path: JsonSerializationHelper) -> g.Model:  # noqa: PLR0912,PLR0911
    """Parse JSON path and return an expression Model."""
    class_name = path._get_class()

    if class_name == typename(g.Sequence):
        items = path._get_array("sequence")
        exprs = [exp_from_json_path(item) for item in items]
        return g.Sequence(sequence=exprs)

    if class_name == typename(g.Choice):
        items = path._get_array("options")
        exprs = [exp_from_json_path(item) for item in items]
        return g.Choice(options=[g.Option(exp=e) for e in exprs])

    if class_name == typename(g.Option):
        exp_path = path._get_nested("exp")
        return g.Option(exp=exp_from_json_path(exp_path))

    if class_name == typename(g.Named):
        name = path._get_string("name")
        exp_path = path._get_nested("exp")
        return g.Named(name=name, exp=exp_from_json_path(exp_path))

    if class_name == typename(g.NamedList):
        name = path._get_string("name")
        exp_path = path._get_nested("exp")
        return g.NamedList(name=name, exp=exp_from_json_path(exp_path))

    if class_name == typename(g.Call):
        name = path._get_string("name")
        return g.Call(name=name)

    if class_name == typename(g.Token):
        token = path._get_string("token")
        return g.Token(token=token)

    if class_name == typename(g.Pattern):
        pattern = path._get_string("pattern")
        return g.Pattern(pattern=pattern)

    if class_name == typename(g.Constant):
        literal = path._opt_str("literal") or ""
        return g.Constant(literal=literal)

    if class_name == typename(g.Alert):
        literal = path._opt_str("literal") or ""
        level = path._opt_int("level") or 0
        return g.Alert(literal=literal, level=level)

    if class_name == typename(g.Group):
        exp_path = path._get_nested("exp")
        return g.Group(exp=exp_from_json_path(exp_path))

    if class_name == typename(g.Optional):
        exp_path = path._get_nested("exp")
        return g.Optional(exp=exp_from_json_path(exp_path))

    if class_name == typename(g.Closure):
        exp_path = path._get_nested("exp")
        return g.Closure(exp=exp_from_json_path(exp_path))

    if class_name == typename(g.PositiveClosure):
        exp_path = path._get_nested("exp")
        return g.PositiveClosure(exp=exp_from_json_path(exp_path))

    if class_name == typename(g.Lookahead):
        exp_path = path._get_nested("exp")
        return g.Lookahead(exp=exp_from_json_path(exp_path))

    if class_name == typename(g.NegativeLookahead):
        exp_path = path._get_nested("exp")
        return g.NegativeLookahead(exp=exp_from_json_path(exp_path))

    if class_name == typename(g.SkipGroup):
        exp_path = path._get_nested("exp")
        return g.SkipGroup(exp=exp_from_json_path(exp_path))

    if class_name == typename(g.SkipTo):
        exp_path = path._get_nested("exp")
        return g.SkipTo(exp=exp_from_json_path(exp_path))

    if class_name == typename(g.Override):
        exp_path = path._get_nested("exp")
        return g.Override(exp=exp_from_json_path(exp_path))

    if class_name == typename(g.OverrideList):
        exp_path = path._get_nested("exp")
        return g.OverrideList(exp=exp_from_json_path(exp_path))

    if class_name == typename(g.Join):
        exp_path = path._get_nested("exp")
        sep_path = path._get_nested("sep")
        return g.Join(
            exp=exp_from_json_path(exp_path),
            sep=exp_from_json_path(sep_path),
        )

    if class_name == typename(g.PositiveJoin):
        exp_path = path._get_nested("exp")
        sep_path = path._get_nested("sep")
        return g.PositiveJoin(
            exp=exp_from_json_path(exp_path),
            sep=exp_from_json_path(sep_path),
        )

    if class_name == typename(g.Gather):
        exp_path = path._get_nested("exp")
        sep_path = path._get_nested("sep")
        return g.Gather(
            exp=exp_from_json_path(exp_path),
            sep=exp_from_json_path(sep_path),
        )

    if class_name == typename(g.PositiveGather):
        exp_path = path._get_nested("exp")
        sep_path = path._get_nested("sep")
        return g.PositiveGather(
            exp=exp_from_json_path(exp_path),
            sep=exp_from_json_path(sep_path),
        )

    if class_name == typename(g.RuleInclude):
        name = path._get_string("name")
        return g.RuleInclude(name=name)

    if class_name == typename(g.Void):
        return g.Void()

    if class_name == typename(g.Cut):
        return g.Cut()

    if class_name == typename(g.EOF):
        return g.EOF()

    if class_name == typename(g.EOL):
        return g.EOL()

    if class_name == typename(g.EmptyClosure):
        return g.EmptyClosure()

    if class_name == typename(g.NameMeta):
        return g.NameMeta()

    if class_name == typename(g.IntMeta):
        return g.IntMeta()

    if class_name == typename(g.UIntMeta):
        return g.UIntMeta()

    if class_name == typename(g.FloatMeta):
        return g.FloatMeta()

    if class_name == typename(g.BoolMeta):
        return g.BoolMeta()

    raise path._error(f"Unsupported: {class_name}")



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""Column identifiers for bar layouts (label, counters, timings)."""

from __future__ import annotations

from enum import StrEnum, auto


__all__ = ["Col"]


class Col(StrEnum):
    # Progress Counters
    begun = auto()
    state = auto()
    pos = auto()
    total = auto()
    label = auto()

    padding = auto()
    percentage = auto()
    pct = auto()

    # Timings & Durations
    elapsed = auto()
    rt = auto()  # Run Time (as timedelta object)
    eta = auto()  # Estimated Time Remaining (as timedelta object)
    eta_s = auto()  # Raw ETA seconds
    # Time components for custom string building
    h = auto()
    m = auto()
    s = auto()
    ms = auto()
    # Core components
    bar = auto()



# Copyright (c) 2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: MIT OR Apache-2.0

"""
json - Direct Value to Grammar translator

This module translates json directly to Grammar,
bypassing the TatSuModel deserializer which fails on modified JSON.
"""

from __future__ import annotations

import json
from typing import Any

from .. import peg as g
from ..util.fromjson import fromjson


def loads_grammar(json_str: str) -> g.Grammar:
    """Parse JSON string and return a Grammar object."""
    value = json.loads(json_str)
    return load_grammar(value)


def load_grammar(value: Any) -> g.Grammar:
    """Parse JSON value and return a Grammar object."""
    result = fromjson(value)
    assert isinstance(result, g.Grammar)
    return result



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Any

from ..contexts import Ctx
from .base import Leaf


class Meta(Leaf):
    pass


class NameMeta(Meta):
    def _parse(self, ctx: Ctx) -> Any:
        return ctx.matchname()

    def _pretty(self, lean: bool = False) -> str:
        _ = lean
        return '@name'


class IntMeta(Meta):
    def _parse(self, ctx: Ctx) -> Any:
        return ctx.matchint()

    def _pretty(self, lean: bool = False) -> str:
        _ = lean
        return '@int'


class UIntMeta(Meta):
    def _parse(self, ctx: Ctx) -> Any:
        return ctx.matchuint()

    def _pretty(self, lean: bool = False) -> str:
        _ = lean
        return '@uint'


class FloatMeta(Meta):
    def _parse(self, ctx: Ctx) -> Any:
        return ctx.matchfloat()

    def _pretty(self, lean: bool = False) -> str:
        _ = lean
        return '@float'


class BoolMeta(Meta):
    def _parse(self, ctx: Ctx) -> Any:
        return ctx.matchbool()

    def _pretty(self, lean: bool = False) -> str:
        _ = lean
        return '@bool'



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import time
from collections.abc import Callable, Generator
from contextlib import contextmanager
from functools import cache
from typing import Any

from ..config import ParserConfig
from ..exceptions import (
    FailedLeftRecursion,
    FailedParse,
    HeartDied,
    ParseException,
)
from ..input import Cursor, NullText, Text
from ..objectmodel import ModelBuilderSemantics
from ..util import (
    prune_dict,
    safe_name,
)
from ..util.boundeddict import BoundedDict
from ..util.heart import Heart
from .ast import AST
from .ctx import Ctx, Func
from .infos import MemoKey, RuleInfo, RuleResult
from .state import ParseState, ParseStateStack
from .tracing import ConsoleTracer, NullTracer, Tracer


type RuleOutcome = RuleResult | ParseException
type MemoCache = dict[MemoKey, RuleOutcome]


@cache
def find_cached_semantic_action(semantics: Any, name: str) -> Callable[..., Any] | None:
    if not semantics:
        return None

    for rulename in (name, safe_name(name), name.strip('_'), f'_{name}', f'_{name}_'):
        action = getattr(semantics, safe_name(rulename), None)
        if callable(action):
            break
    else:
        action = None

    if not callable(action):
        action = getattr(semantics, '_default', None)

    if not callable(action):
        action = None

    return action


class ParserCore(Ctx):
    states: ParseStateStack  # type: ignore

    def __init__(
        self,
        /,
        *,
        config: ParserConfig | None = None,
        asmodel: bool = False,
        **settings: Any,
    ) -> None:
        super().__init__()

        config = ParserConfig.new(config, **settings)
        assert isinstance(config, ParserConfig)
        assert isinstance(config, ParserConfig)
        self._config: ParserConfig = config
        self._active_config: ParserConfig = self._config

        self.input: Text = NullText()
        self.states: ParseStateStack = ParseStateStack(cursor=self.input.newcursor())
        if not self.config.semantics and asmodel:
            self.config.semantics = ModelBuilderSemantics()
        self.semantics: type | None = config.semantics
        self._furthest_exception: FailedParse | None = None

        self._initialize_caches()
        self.tracer: Tracer = NullTracer()
        self.heart: Heart | None = config.heart
        self.lastbeat_time = 0.0
        self.lastbeat_pos: int = 0
        self.update_tracer()

    def _initialize_caches(self) -> None:
        self._furthest_exception = None
        self._memos: MemoCache = BoundedDict(
            int(max(1.0, self.config.perlinememos) * self.cursor.linecount)
        )
        self._results: MemoCache = {}
        self.states = ParseStateStack(cursor=self.input.newcursor())

    def _reset(self) -> None:
        self._initialize_caches()
        self.keywords: set[str] = set(self.config.keywords or ())
        self.semantics = self.config.semantics
        if self.semantics and hasattr(self.semantics, 'set_context'):
            self.semantics.set_context(self)

    @property
    def config(self):
        return self._active_config

    @property
    def active_config(self) -> ParserConfig:
        return self._active_config

    @property
    def self_config(self) -> ParserConfig:
        return self._config

    @property
    def cursor(self) -> Cursor:
        return self.state.cursor

    @property
    def pos(self) -> int:
        return self.states.state.cursor.pos

    @property
    def line(self) -> int:
        return self.cursor.line

    @property
    def state(self) -> ParseState:
        return self.states.state

    @property
    def ast(self) -> AST:
        return self.states.state.ast

    @ast.setter
    def ast(self, value: AST) -> None:
        self.states.state.ast = value

    @property
    def cst(self) -> Any:
        return self.states.state.cst

    @cst.setter
    def cst(self, value: Any) -> None:
        self.states.state.cst = value

    @property
    def callstack(self) -> list[RuleInfo]:
        return self.states.callstack

    def update_tracer(self) -> Tracer:
        if self.active_config.trace:
            tracer: Tracer = ConsoleTracer(config=self.config)
        else:
            tracer = NullTracer()
        self.tracer = tracer
        return self.tracer

    def set_furthest_exception(self, e: FailedParse) -> None:
        if not self._furthest_exception or e.pos >= self._furthest_exception.pos:
            self._furthest_exception = e

    def goto(self, pos: int) -> None:
        self.states.state.cursor.goto(pos)

    def _next(self) -> Any:
        return self.cursor.next()

    def heartbeat(self) -> bool:
        if self.heart is None:
            return False

        if self.heart.dead():
            raise HeartDied("Heart is dead")

        now = time.perf_counter()
        if (now - self.lastbeat_time) <= (1.0 / self.config.heart_bps):
            return False

        if self.pos <= self.lastbeat_pos:
            return False

        try:
            self.heart.beat(self.cursor.line, self.cursor.linecount)
        finally:
            self.lastbeat_time = now
            self.lastbeat_pos = self.pos
        return True

    def next_token(self, ri: RuleInfo | None = None) -> None:
        if not (ri and ri.is_tokn):
            self.state.cursor.next_token()

    def _define(self, keys: list[str], addkeys: list[str] | None = None) -> None:
        # NOTE: called by generated parsers
        return self.state.define(keys, addkeys)

    def define(self, keys: list[str], addkeys: list[str] | None = None) -> None:
        # NOTE: called by generated parsers
        return self.state.define(keys, addkeys)

    def setname(self, name: str) -> None:
        # NOTE: called by generated parsers
        self.state.nameset(name)

    def addname(self, name: str) -> None:
        # NOTE: called by generated parsers
        self.state.nameadd(name)

    @contextmanager
    def statescope(self, *, merge: bool = True) -> Generator[ParseState, None, None]:
        self.states.push()
        try:
            yield self.state
            if merge:
                self.states.merge()
            else:
                self.states.pop()
        except FailedParse:
            self.states.undo()
            raise

    def cut(self) -> None:
        self.state.cutseen = True
        self.tracer.trace_cut(self)

        if not self.config.prune_memos_on_cut:
            return

        cutpos = self.pos

        def unwanted(key: MemoKey, value: RuleResult | ParseException) -> bool:
            return key.pos < cutpos and not isinstance(value, FailedLeftRecursion)

        prune_dict(self._memos, unwanted)

    _cut = cut

    def find_rule(self, name: str) -> Func:
        assert name
        raise NotImplementedError

    def find_semantic_action(self, name: str) -> Callable[..., Any] | None:
        return find_cached_semantic_action(self.semantics, name)

    def newexcept(
        self,
        msg: Any,
        excls: type[FailedParse] = FailedParse,
    ) -> FailedParse:
        return excls(self.cursor, self.callstack, msg)

    @property
    def ruleinfo(self) -> RuleInfo:
        return self.callstack[-1]

    def memokey(self) -> MemoKey:
        return MemoKey(self.pos, self.ruleinfo)

    def memo(self, key: MemoKey) -> RuleOutcome | None:
        return self._memos.get(key)

    def memoize(
        self,
        key: MemoKey,
        memo: RuleResult | ParseException,
    ) -> RuleResult | ParseException:
        if key.ruleinfo.memoizable and self.config.memoization:
            self._memos[key] = memo
        return memo



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import functools
from collections.abc import Callable
from typing import Any

from ..ctx import Ctx, Func
from ..infos import RuleInfo


def tatsumasu(*params: Any, **kwparams: Any) -> Callable[[Callable[[], None]], Func]:
    def decorator(func: Callable[[], None]) -> Func:
        @functools.wraps(func)
        def wrapper(self: Ctx, _ctx: Ctx | None = None) -> Any:
            ruleinfo = RuleInfo.new(self, func, params, kwparams)
            return self.call(ruleinfo)

        return wrapper

    return decorator  # pyright: ignore[reportReturnType]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from copy import copy
from dataclasses import field
from functools import cached_property
from typing import Any, Self, cast

from ..contexts import AST, Ctx
from ..contexts.cst import cstmerge
from ..exceptions import FailedParse, FailedRef
from ..objectmodel import nodedataclass
from ..util import indent, trim, typename
from .base import PEP8_LLEN, Box, Leaf, Model, Rule
from .math import ffset, kdot, ref


@nodedataclass
class Group(Box):
    def _parse(self, ctx: Ctx) -> Any:
        return self.exp._parse(ctx)

    def _pretty(self, lean=False):
        exp = self.exp._pretty(lean=lean)
        if len(exp.splitlines()) <= 1:
            return f'({trim(exp)})'
        return f'(\n{indent(exp)}\n)'

    def optimized(self) -> Model:
        newexp = self.exp.optimized()
        assert isinstance(newexp, Model)

        if isinstance(newexp, Leaf | Group):
            return newexp
        return Group(newexp)


@nodedataclass
class SkipGroup(Box):
    def _parse(self, ctx: Ctx) -> Any:
        with ctx.skipgroup():
            self.exp._parse(ctx)
            return None

    def _pretty(self, lean=False):
        exp = self.exp._pretty(lean=lean)
        if len(exp.splitlines()) <= 1:
            return f'(?:{trim(exp)})'
        return f'(?:\n{indent(exp)}\n)'


@nodedataclass
class Lookahead(Box):
    def _parse(self, ctx: Ctx) -> Any:
        with ctx.if_():
            return self.exp._parse(ctx)

    def _pretty(self, lean=False):
        return '&' + self.exp._pretty(lean=lean)

    @cached_property
    def _nullable(self) -> bool:
        return True


@nodedataclass
class NegativeLookahead(Box):
    def _parse(self, ctx: Ctx) -> Any:
        with ctx.ifnot_():
            return self.exp._parse(ctx)

    def _pretty(self, lean=False):
        return '!' + str(self.exp._pretty(lean=lean))

    @cached_property
    def _nullable(self) -> bool:
        return True


@nodedataclass
class SkipTo(Box):
    def _parse(self, ctx: Ctx) -> Any:
        return ctx.skip_to(self.exp._parse)

    def _first(self, k, f) -> ffset:
        return {('.',)} | super()._first(k, f)

    def _pretty(self, lean=False):
        return '->' + self.exp._pretty(lean=lean)


@nodedataclass
class Optional(Box):
    def _parse(self, ctx: Ctx) -> Any:
        ctx.states.push()
        try:
            self._add_defined(ctx)
            value = self.exp._parse(ctx)
            ctx.states.merge()
            return value
        except FailedParse:
            if ctx.states.undo().cutseen:
                raise
            return None

    def _first(self, k, f) -> ffset:
        return {()} | self.exp._first(k, f)

    def _pretty(self, lean=False):
        exp = self.exp._pretty(lean=lean)
        if len(exp.splitlines()) <= 1:
            return f'[{exp}]'
        return f'[\n{exp}\n]'

    @cached_property
    def _nullable(self) -> bool:
        return True

    def optimized(self) -> Self | Model:
        from .closure import Closure, Gather, Join

        exp = self.exp.optimized()
        if isinstance(
            exp, Optional | Closure | Join | Gather
        ) and 'Positive' not in typename(exp):
            return exp
        new = copy(self)
        new.exp = exp
        return new


@nodedataclass
class Sequence(Model):
    sequence: list[Model] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        if not self.sequence:
            self.sequence = list(self.ast or [])
        assert isinstance(self.sequence, list), self.sequence

    def _parse(self, ctx: Ctx) -> Any:
        self._add_defined(ctx)
        out = None
        for s in self.sequence:
            exp = s
            while isinstance(exp, Group):
                exp = exp.exp

            r = exp._parse(ctx)
            if r is None:
                continue

            out = cstmerge(out, r)
        return out

    @cached_property
    def defines_single(self) -> list[str]:
        return list(set().union(*(s.defines_single for s in self.sequence)))

    @cached_property
    def defines_list(self) -> list[str]:
        return list(set().union(*(s.defines_list for s in self.sequence)))

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        return set().union(*(s.missing_rules(rulenames) for s in self.sequence))

    def _used_rule_names(self):
        return set().union(*(s._used_rule_names() for s in self.sequence))

    def _first(self, k, f) -> ffset:
        result: ffset = {()}
        for s in self.sequence:
            assert isinstance(s, Model), f'{type(s)}:{s} is not a Model'
            x = s._first(k, f)
            result = kdot(result, x, k)
        self._firstset = result
        return result

    def _follow(self, k, fl, a):
        fs = a
        for x in reversed(self.sequence):
            if isinstance(x, Call):
                fl[x.name] |= fs
            x._follow(k, fl, fs)
            fs = kdot(x.firstset(k=k), fs, k)
        return a

    def nodecount(self) -> int:
        return 1 + sum(s.nodecount() for s in self.sequence)

    def _pretty(self, lean=False):
        seq = [str(s._pretty(lean=lean)) for s in self.sequence]
        single = ' '.join(seq)
        if len(single) <= PEP8_LLEN and len(single.splitlines()) <= 1:
            return single
        else:
            return '\n'.join(seq)

    @cached_property
    def _nullable(self) -> bool:
        return all(s._nullable for s in self.sequence)

    def optimized(self) -> Model | Self:
        seq = [e.optimized() if isinstance(e, Model) else e for e in self.sequence]
        for s in seq:
            assert isinstance(s, Model)
        if len(seq) == 1 and isinstance(seq[0], Model):
            return seq[0]
        new = copy(self)
        new.sequence = seq
        return new


@nodedataclass
class Call(Leaf):
    name: str = ''
    _rule: Rule | None = None

    def __post_init__(self):
        if not self.name:
            self.ast = AST(name=self.ast)
        super().__post_init__()
        assert isinstance(self.name, str), self.name

    @property
    def rule(self) -> Rule | None:
        return self._rule

    def follow_ref(self) -> Model:
        return self.grammar.rulemap.get(self.name, self)

    def _parse(self, ctx: Ctx) -> Any:
        try:
            if self._rule:
                return ctx.expcall(self._rule._parse)
            parse = ctx.find_rule(self.name)
            return ctx.expcall(parse)
        except KeyError as e:
            raise ctx.newexcept(self.name, excls=FailedRef) from e

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        if self.name not in rulenames:
            return {self.name}
        return set()

    def _used_rule_names(self) -> set[str]:
        return {self.name}

    def _first(self, k, f) -> ffset:
        _ = k
        return f[self.name] | {ref(self.name)}

    def _follow(self, k, fl, a):
        _ = k
        fl[self.name] |= a
        return set(a) | {self.name}

    def _pretty(self, lean=False):
        _ = lean
        return self.name

    def is_nullable(self) -> bool:
        return self.grammar.rulemap[self.name]._nullable

    def optimized(self) -> Call:
        if not self._rule or not isinstance(self._rule.exp, Call):
            return self

        rule = cast(Rule, self._rule)
        assert rule is not None
        while isinstance(rule.exp, Call):
            call = rule.exp
            if not call._rule:
                break
            rule = call._rule
        new = Call(
            name=rule.name,
        )
        new._rule = rule
        return new

    def __str__(self):
        return str(ref(self.name))



#  Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
#  SPDX-License-Identifier: BSD-4-Clause
#

# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from functools import cached_property
from typing import Any

from ..contexts import _AT_, AST, Ctx
from ..objectmodel import nodedataclass
from ..util import typename
from .base import Box, NamedBox


@nodedataclass
class Named(NamedBox):
    def __post_init__(self):
        if not self.ast:
            self.ast = AST(name=self.name, exp=self.exp)
        super().__post_init__()
        if not self.name:
            raise TypeError(f"{typename(self)} is missing required: 'name'")
        assert getattr(self, 'name', None) is not None

    def _parse(self, ctx: Ctx) -> Any:
        value = self.exp._parse(ctx)
        ctx.ast[self.name] = value
        return value

    @cached_property
    def defines_single(self) -> list[str]:
        return list({self.name, *super().defines_single})

    def _pretty(self, lean=False):
        if lean:
            return self.exp._pretty(lean=True)
        return f'{self.name}={self.exp._pretty(lean=lean)}'


@nodedataclass
class NamedList(Named):
    def _parse(self, ctx: Ctx) -> Any:
        value = self.exp._parse(ctx)
        ctx.ast._setlist(self.name, value)
        return value

    @cached_property
    def defines_list(self) -> list[str]:
        return list({self.name, *super().defines_list})

    def _pretty(self, lean=False):
        if lean:
            return self.exp._pretty(lean=True)
        return f'{self.name}+={self.exp._pretty(lean=lean)!s}'


@nodedataclass
class Override(Box):
    def _parse(self, ctx: Ctx) -> Any:
        value = self.exp._parse(ctx)
        ctx.ast[_AT_] = value
        return {_AT_: value}

    def _pretty(self, lean=False):
        if lean:
            return self.exp._pretty(lean=True)
        return f'={self.exp._pretty(lean=lean)!s}'


@nodedataclass
class OverrideList(Box):
    def _parse(self, ctx: Ctx) -> Any:
        value = self.exp._parse(ctx)
        if _AT_ not in ctx.ast:
            value = [value]
        ctx.ast[_AT_] = value
        return {_AT_: value}

    def _pretty(self, lean=False):
        if lean:
            return self.exp._pretty(lean=True)
        return f'+={self.exp._pretty(lean=lean)!s}'



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""Approximate benchmarks for the file-backed packet queue.

Runs with ``python -m tatsu.packetz.test_packetz``.
Also compatible with ``pytest`` if available.
"""

from __future__ import annotations

import time
from pathlib import Path
from typing import Any

from .packet import Packet, PacketLike, pack
from .queue import PacketzQueue


QUEUE_PATH = Path("test_packetz.jsonl")


# ---------------------------------------------------------------------------
# helpers
# ---------------------------------------------------------------------------


def _new_queue():
    QUEUE_PATH.unlink(missing_ok=True)
    return PacketzQueue(path=QUEUE_PATH)


def _make_payload(n: int) -> str:
    if n == 0:
        return ""
    chunk = "The quick brown fox jumps over the lazy dog. " * ((n // 44) + 1)
    return chunk[:n]


_SENTINEL = object()


def _write_batch(
    q: PacketzQueue, n: int, payload: Any = _SENTINEL
) -> tuple[float, int]:
    start = time.perf_counter()
    for _ in range(n):
        if payload is _SENTINEL:
            q.send()
        else:
            q.send(data=payload)
    elapsed = time.perf_counter() - start
    size = QUEUE_PATH.stat().st_size
    return elapsed, size


def _read_all(q: PacketzQueue) -> tuple[list[PacketLike], float]:
    start = time.perf_counter()
    packets = list(q.receive())
    elapsed = time.perf_counter() - start
    return packets, elapsed


# ---------------------------------------------------------------------------
# benchmark cases
# ---------------------------------------------------------------------------


def bench_write_small():
    q = _new_queue()
    for n in (1, 10, 100, 1000):
        elapsed, size = _write_batch(q, n)
        rate = f"{n / elapsed:.0f}/s" if elapsed else "∞"
        print(f"  write {n:>5} pkts  {elapsed:.3f}s  {rate:>8}  file={size:,}B")


def bench_write_payload_size():
    n = 100
    for label, kb in (("empty", 0), ("1KB", 1), ("10KB", 10), ("100KB", 100)):
        q = _new_queue()
        payload = _make_payload(kb * 1024)
        elapsed, size = _write_batch(q, n, payload=payload)
        rate = f"{n / elapsed:.0f}/s" if elapsed else "∞"
        print(f"  {label:>6}  {n:>3} pkts  {elapsed:.3f}s  {rate:>8}  file={size:,}B")


def bench_read():
    for label, n in (
        ("1 pkt", 1),
        ("10 pkts", 10),
        ("100 pkts", 100),
        ("1000 pkts", 1000),
    ):
        q = _new_queue()
        _write_batch(q, n, payload=_make_payload(256))
        _packets, elapsed = _read_all(q)
        rate = f"{n / elapsed:.0f}/s" if elapsed else "∞"
        print(f"  read {label:>10}  {elapsed:.4f}s  {rate:>8}")


def bench_growth_penalty():
    q = _new_queue()
    payload = _make_payload(256)
    for round_ in range(4):
        _write_batch(q, 500, payload=payload)
        packets, elapsed = _read_all(q)
        size = QUEUE_PATH.stat().st_size
        rate = f"{len(packets) / elapsed:.0f}/s" if elapsed else "∞"
        print(
            f"  growth round {round_}: file={size:,}B  read {len(packets)} pkts @ {rate}"
        )


def bench_round_trip():
    q = _new_queue()
    payloads = [
        None,
        "",
        42,
        3.14,
        "hello world",
        _make_payload(1024),
        {"nested": ["list", {"key": "value"}]},
    ]
    sent_ids = []
    for p in payloads:
        pkt = q.send(data=p, to="bench")
        sent_ids.append(pkt.id)
    packets, elapsed = _read_all(q)
    ok = [p.id for p in packets] == sent_ids
    print(
        f"  sent {len(sent_ids)} pkts, received {len(packets)} in {elapsed:.3f}s  ids_match={ok}"
    )


def bench_dedup():
    q = _new_queue()
    pkt = Packet(data="dedup-test", to="check")
    raw = pack(pkt) + "\n"
    for _ in range(2):
        with QUEUE_PATH.open("at", encoding="utf-8") as f:
            f.write(raw)
    packets, elapsed = _read_all(q)
    ok = len(packets) == 1
    print(
        f"  wrote 2 identical lines, got {len(packets)} unique ({'PASS' if ok else 'FAIL'}) in {elapsed:.3f}s"
    )


def bench_transforms_wire_size():
    """Compare wire size: raw JSON vs. current pack() transforms."""
    import json

    from ..util.asjson import asjson
    from .packet import Packet, pack

    pkt = Packet(data=_make_payload(256), to="size-test")
    raw = json.dumps(asjson(pkt), separators=(",", ":"), ensure_ascii=False)
    cooked = pack(pkt)
    ratio = f"{len(cooked) / len(raw) * 100:.0f}%" if raw else "N/A"
    print(f"  raw JSON        {len(raw):>6}B")
    print(f"  after pack()    {len(cooked):>6}B  ({ratio})")


def bench_bloat():
    cases = [
        ("empty", ""),
        ("5KB_text", "x" * 5000),
        ("5KB_spaces", " " * 5000),
        ("4KB_repeat", "x y z " * 1000),
    ]
    n = 100
    for label, payload in cases:
        q = _new_queue()
        _elapsed, size = _write_batch(q, n, payload=payload)
        raw_data_size = n * (len(payload) if payload else 0)
        ratio = f"{size / max(raw_data_size, 1):.1f}x"
        print(
            f"  {label:>12}  {n:>3} pkts  wire={size:,}B  raw-data={raw_data_size:,}B  bloat={ratio}"
        )


# ---------------------------------------------------------------------------
# runner
# ---------------------------------------------------------------------------


def main():
    BENCHMARKS = [
        ("Write throughput (small payload)", bench_write_small),
        ("Write throughput (varying payload size)", bench_write_payload_size),
        ("Read throughput", bench_read),
        ("Queue growth penalty", bench_growth_penalty),
        ("Round-trip value fidelity", bench_round_trip),
        ("Duplicate dedup", bench_dedup),
        ("Serialization transforms wire size", bench_transforms_wire_size),
        ("Serialization bloat factor", bench_bloat),
    ]
    print(f"packetz queue: {QUEUE_PATH}")
    print()
    for title, fn in BENCHMARKS:
        print(f"[{title}]")
        try:
            fn()
        except Exception as e:
            print(f"  ERROR: {e}")
        print()
    # cleanup
    QUEUE_PATH.unlink(missing_ok=True)


if __name__ == "__main__":
    main()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .choice import ChoiceContext
from .exp import ExpContext
from .expsep import ExpWithSepContext
from .loop import LoopContext
from .loopsep import LoopWithSepContext


__all__ = [
    'ChoiceContext',
    'ExpContext',
    'ExpWithSepContext',
    'LoopContext',
    'LoopWithSepContext',
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re
from functools import cached_property
from typing import Any, Self

from ..config import ParserConfig
from ..util import (
    Undefined,
    linecount,
    notnone,
    str_from_match,
    typename,
)
from ..util.newlines import take_linebreak_len, take_non_newline_whitespace_len
from ..util.regextools import cached_re_compile
from .cursor import Cursor, Text, matchbool, matchfloat, matchint, matchname, matchuint
from .infos import LineIndexInfo, LineInfo, PosLine


DEFAULT_WHITESPACE_RE = re.compile(r'(?m)\s+')


class TextLinesCursor(Cursor):
    __slots__ = ('_input', 'len', 'pos', 'textstr')

    def __init__(self, input: TextLines, pos: int = 0):
        self._input: TextLines = input
        self.pos: int = pos
        self.len: int = input.len
        self.textstr = input.textstr
        self._namechars = input._namechar_set

    def clone(self) -> Self:
        return type(self)(self.input, pos=self.pos)

    def __copy__(self) -> Self:
        return type(self)(self.input, pos=self.pos)

    @property
    def input(self) -> TextLines:
        return self._input

    @property
    def line(self) -> int:
        return self.lineat(self.pos)

    @property
    def linecount(self) -> int:
        return self.input.linecount

    @property
    def col(self) -> int:
        return self.poscol(self.pos)

    @property
    def source(self) -> str:
        n = min(len(self.input.line_index) - 1, self.line)
        source, _actual_line = self.input.line_index[n]
        return source

    @cached_property
    def namechars(self) -> set[str]:
        return self._namechars

    def goto(self, pos: int):
        self.pos = max(0, min(self.len, pos))

    def move(self, n: int):
        self.goto(self.pos + n)

    def atend(self) -> bool:
        return self.pos >= self.len

    def ateol(self) -> bool:
        return self.atend() or self.current in {'\r', '\n', None}

    @property
    def current(self) -> str | None:
        if self.pos >= self.len:
            return None
        return self.textstr[self.pos]

    def peek(self, n: int = 1) -> str | None:
        p = self.pos + n
        if p >= self.len or p < 0:
            return None
        return self.textstr[p]

    def next(self) -> str | None:
        if self.atend():
            return None
        c = self.current
        self.move(1)
        return c

    def next_token(self) -> None:
        p = -1
        while self.pos != p:
            p = self.pos
            self.eat_whitespace()
            while self.eat_eol_comments():
                self.eat_whitespace()
            self.eat_comments()

    def eat_spaces_no_newlines(self):
        p = None
        while self.pos != p:
            p = self.pos
            self.pos += take_non_newline_whitespace_len(self.textstr, self.pos)
            if self.eat_eol_comments():
                self.pos += take_non_newline_whitespace_len(self.textstr, self.pos)
            self.eat_comments()

    def matcheol(self) -> bool:
        mark = self.pos
        self.eat_spaces_no_newlines()
        eol_len = take_linebreak_len(self.textstr, self.pos)
        if eol_len is None:
            self.pos = mark
            return False
        self.move(eol_len)
        self.eat_spaces_no_newlines()
        return True

    def match(self, token: str) -> str | None:
        if not token:
            return None

        p = self.pos
        text = self.textstr[p : p + len(token)]

        if self.input.ignorecase:
            is_match = text.lower() == token.lower()
        else:
            is_match = text == token

        if not is_match:
            return None

        self.move(len(token))
        partial_match = (
            self.input.nameguard
            and self.is_name_char(self.current)
            and self.is_name(token)
        )
        if partial_match:
            self.goto(p)
            return None

        return token

    def matchre(self, pattern: str | re.Pattern) -> str | None:
        if not (match := self._scanre(pattern)):
            return None
        matched = match.group()
        token = str_from_match(match)
        self.move(len(matched))
        return token

    def matchname(self) -> str | None:
        return matchname(self)

    def matchint(self) -> int | None:
        return matchint(self)

    def matchuint(self) -> int | None:
        return matchuint(self)

    def matchfloat(self) -> float | None:
        return matchfloat(self)

    def matchbool(self) -> bool | None:
        return matchbool(self)

    def is_name_char(self, c: str | None) -> bool:
        return c is not None and (c.isalnum() or c in self.namechars)

    def is_name(self, s: str) -> bool:
        if not s:
            return False

        goodstart = s[0].isalpha() or s[0] in self.namechars
        return goodstart and all(self.is_name_char(c) for c in s[1:])

    def lineinfo(self, pos: int | None = None) -> LineInfo:
        pos = notnone(pos, self.pos) or 0
        input = self.input

        if not input.line_cache or not input.line_index:
            return LineInfo(
                source=input.source,
                line=0,
                col=0,
                start=0,
                end=input.len,
                text=input.textstr,
            )

        # Ensure start is within bounds for cache lookup
        # The cache has an extra entry at the end, so len - 2 is the last valid index for content
        pos = min(pos, len(input.line_cache) - 2)
        start, line, length = input.line_cache[pos]
        end = start + length
        col = pos - start
        text = input.textstr[start:end]

        n = min(len(input.line_index) - 1, line)
        source, actual_line = input.line_index[n]
        return LineInfo(
            source=source,
            line=actual_line,
            col=col,
            start=start,
            end=end,
            text=text,
        )

    def lookahead_pos(self) -> str:
        if self.atend():
            return ''
        info = self.lineinfo(self.pos)
        return '[%d:%d]' % (info.line + 1, info.col + 1)

    def lookahead(self) -> str:
        if self.atend():
            return ''
        info = self.lineinfo(self.pos)
        text = info.text[info.col : info.col + 1 + 80]
        return self.input.split_block_lines(text)[0].rstrip()

    def lineat(self, pos: int | None = None) -> int:
        pos = notnone(pos, self.pos) or 0
        if not self.input.line_cache:
            return 0
        return self.input.line_cache[pos].lineno

    def poscol(self, pos: int | None = None) -> int:
        pos = notnone(pos, self.pos) or 0
        if not self.input.line_cache:
            return 0
        start = self.input.line_cache[pos].startpos
        return pos - start

    def get_line(self, n: int | None = None) -> str:
        return self.input.get_line(notnone(n, self.line))

    def get_lines(
        self,
        start: int | None = None,
        end: int | None = None,
    ) -> list[str]:
        return self.input.get_lines(start, end)

    def line_index(self, start: int = 0, end: int | None = None) -> list[LineIndexInfo]:
        return self.input.line_index_at(start, end)

    def eat_whitespace(self) -> bool:
        if self.input.whitespace_re:
            return self._eat_regex(self.input.whitespace_re)
        return False

    def eat_comments(self) -> bool:
        return self._eat_regex(self.input.config.comments)

    def eat_eol_comments(self) -> bool:
        return self._eat_regex(self.input.config.eol_comments)

    def _eat_regex(self, regex: str | re.Pattern | None) -> bool:
        if not regex:
            return False
        res = False
        while self._matchre_fast(regex):
            res = True
        return res

    def _matchre_fast(self, pattern: str | re.Pattern | None) -> bool:
        if not (match := self._scanre(pattern)):
            return False
        self.goto(match.end())
        return True

    def _scanre(self, pattern: str | re.Pattern | None) -> re.Match[Any] | None:
        if not (cre := cached_re_compile(pattern)):
            return None
        return cre.match(self.textstr, self.pos)

    def __len__(self) -> int:
        return self.len

    def __repr__(self) -> str:
        pos = self.pos
        return f'{typename(self)}({pos=})'


class TextLines(Text):
    def __init__(
        self,
        text: str,
        *,
        config: ParserConfig | None = None,
        **settings: Any,
    ):
        config = ParserConfig.new(config=config, **settings)
        assert isinstance(config, ParserConfig)
        self.config = config

        text = str(text)
        self.original_text = text

        self.whitespace_re = self.build_whitespace_re(config.whitespace)
        self.nameguard = (
            config.nameguard
            if config.nameguard is not None
            else bool(self.whitespace_re) or bool(config.namechars)
        )
        self._namechar_set = set(config.namechars or '')

        # Structural data
        self.textstr = ""
        self.lines: list[str] = []
        self.line_index: list[LineIndexInfo] = []
        self.line_cache: list[PosLine] = []

        self._preprocess()
        self._postprocess()

    def newcursor(self) -> Cursor:
        return TextLinesCursor(self)

    @property
    def source(self) -> str:
        return str(self.config.source or '')

    @property
    def len(self) -> int:
        return len(self.textstr)

    @property
    def linecount(self) -> int:
        return linecount(self.textstr)

    @property
    def ignorecase(self) -> bool:
        return bool(self.config.ignorecase)

    @staticmethod
    def build_whitespace_re(whitespace: Any) -> re.Pattern | None:
        if whitespace is Undefined:
            return DEFAULT_WHITESPACE_RE
        if whitespace in {None, ''}:
            return None
        if isinstance(whitespace, re.Pattern):
            return whitespace
        if whitespace:
            return cached_re_compile(whitespace)
        return None

    def _preprocess(self):
        lines, index = self._preprocess_block(self.source, self.original_text)
        self.lines = lines
        self.line_index = index
        self.textstr = self.join_block_lines(lines)

    def _postprocess(self):
        cache, _count = PosLine.build_line_cache(self.lines, len(self.textstr))
        self.line_cache = cache

    def _preprocess_block(
        self,
        name: str,
        block: str,
        **kwargs,
    ) -> tuple[list[str], list[LineIndexInfo]]:
        lines = self.split_block_lines(block)
        index = LineIndexInfo.block_index(name, len(lines))
        return self.process_block(name, lines, index, **kwargs)

    def split_block_lines(self, block: str) -> list[str]:
        return block.splitlines(True)

    def join_block_lines(self, lines: list[str]) -> str:
        return ''.join(lines)

    def process_block(
        self,
        _name: str,
        lines: list[str],
        index: list[LineIndexInfo],
        **_kwargs,
    ):
        return lines, index

    def get_line(self, n: int) -> str:
        return self.lines[n]

    def get_lines(self, start: int | None = None, end: int | None = None) -> list[str]:
        if start is None:
            start = 0
        return self.lines[start : notnone(end, start + 1)]

    def line_index_at(
        self, start: int = 0, end: int | None = None
    ) -> list[LineIndexInfo]:
        if end is None:
            end = len(self.line_index)
        return self.line_index[start : 1 + end]


TextLinesTokenizer = TextLines



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import ast as stdlib_ast
import inspect
from collections.abc import Generator
from contextlib import contextmanager, suppress
from typing import Any

from ..config import ParserConfig
from ..exceptions import (
    FailedLeftRecursion,
    FailedParse,
    FailedSemantics,
    KeywordError,
    ParseException,
)
from ..input import NullCursor, NullText, Text
from ..input.textlines import TextLines
from ..objectmodel import ModelBuilderSemantics
from ..util import (
    Undefined,
    boundcall,
    is_eval_safe,
    prune_dict,
    safe_builtins,
    safe_eval,
    trim,
)
from .ast import AST
from .core import ParserCore
from .cst import closedlist, islist
from .ctx import CanParse, Ctx, is_func
from .infos import MemoKey, ParseInfo, RuleInfo, RuleResult
from .state import ParseStateStack


type RuleOutcome = RuleResult | ParseException
type MemoCache = dict[MemoKey, RuleOutcome]


class ParserEngine(ParserCore, CanParse):
    def parse(
        self,
        text: str | Text,
        /,
        *,
        start: str | None = None,
        config: Any = None,
        asmodel: bool = False,
        **settings: Any,
    ) -> Any:
        with self.bound(
            text,
            start=start,
            config=config,
            asmodel=asmodel,
            **settings,
        ):
            actual_start: str = self.config.effective_start_rule_name() or 'start'
            rule = self.find_rule(actual_start)
            return rule(self)

    @contextmanager
    def bound(
        self,
        text: str | Text,
        /,
        *,
        start: str | None = None,
        config: Any = None,
        asmodel: bool = False,
        **settings: Any,
    ) -> Generator[Ctx, None, None]:
        config = self.config.override_config(config)
        assert isinstance(config, ParserConfig)
        config = config.override(start=start, **settings)
        assert isinstance(config, ParserConfig)
        self._active_config = config

        self._initialize_caches()
        self.heart = config.heart
        self.lastbeat_time = 0.0
        self.lastbeat_pos: int = 0
        self._furthest_exception = None
        self.update_tracer()
        try:
            if isinstance(text, Text):
                input = text
            else:
                input = TextLines(text=text, config=config)
            assert not isinstance(input, NullText)
            self.input = input
            self.states = ParseStateStack(cursor=input.newcursor())
            assert not isinstance(self.state.cursor, NullCursor)
            self._reset()

            if not self.config.semantics and asmodel:
                self.config.semantics = ModelBuilderSemantics()
            if self.config.semantics and hasattr(self.config.semantics, 'set_context'):
                self.config.semantics.set_context(self)  # ty: ignore[call-non-callable]

            yield self

        except FailedParse as e:
            self.set_furthest_exception(e)
            if isinstance(self._furthest_exception, Exception):
                raise self._furthest_exception from e
            raise
        finally:
            self._initialize_caches()
            self._active_config = self._config
            self.update_tracer()
            if self.config.semantics and hasattr(self.config.semantics, 'set_context'):
                self.config.semantics.set_context(None)  # ty: ignore[call-non-callable]

    def call(self, ri: RuleInfo) -> Any:
        self.heartbeat()
        if ri.should_trace:
            self.callstack.append(ri)
        self.next_token(ri)
        key = self.memokey()

        pos = self.pos
        try:
            self.tracer.trace_entry(self)

            if ri.is_lrec:
                result = self.recursive_call(ri, key)
            else:
                result = self.rule_call(ri, key)

            self.goto(result.newpos)
            self.state.append(result.node)

            self.tracer.trace_success(self)

            return result.node
        except FailedParse as e:
            self.goto(pos)
            self.set_furthest_exception(e)
            self.tracer.trace_failure(self, e)
            raise
        finally:
            if ri.should_trace:
                self.callstack.pop()

    def recursive_call(self, ri: RuleInfo, key: MemoKey) -> RuleResult:
        if not ri.is_lrec:
            return self.rule_call(ri, key)
        elif not self.config.left_recursion:
            raise self.newexcept('Left recursion detected', excls=FailedLeftRecursion)

        result: RuleResult | ParseException | None = self._results.get(key)
        if isinstance(result, RuleResult):
            return result
        elif isinstance(result, Exception):
            raise result

        result = self.newexcept(ri.name, FailedLeftRecursion)
        assert isinstance(result, RuleResult | ParseException)
        self._results[key] = result

        initial = self.pos
        lastpos = -1
        while True:
            self.clear_recursion_errors()
            try:
                new_result = self.rule_call(ri, key)
                self.goto(initial)
            except FailedParse:
                break

            if new_result.newpos > lastpos:
                self.save_result(key, new_result)
                lastpos = new_result.newpos
                result = new_result
            else:
                break

        if isinstance(result, Exception):
            raise result

        assert isinstance(result, RuleResult | ParseException)
        return result

    def rule_call(self, ri: RuleInfo, key: MemoKey) -> RuleResult:
        result = self.memo(key)
        if isinstance(result, Exception):
            raise result
        if isinstance(result, RuleResult):
            return result

        self.set_left_recursion_guard(key)

        self.states.new()
        try:
            self.next_token(ri)

            node = self.func_call(ri)
            node = self.semantics_call(ri, node, pos=key.pos)
            self.set_parseinfo(node, ri.name, key.pos)

            result = RuleResult(node, self.pos)
            self.memoize(key, result)

            return result
        except FailedSemantics as e:
            ex = self.newexcept(str(e))
            self.memoize(key, ex)
            raise ex from e
        except ParseException as e:
            self.memoize(key, e)
            raise
        finally:
            self.states.undo()

    def func_call(self, ri: RuleInfo) -> Any:
        with self.statescope():
            is_legacy_parser = ri.instance is self
            if is_legacy_parser:
                ri.func(ri.instance)
            elif inspect.ismethod(ri.func) or is_func(ri.func):
                ri.func(self)
            else:
                ri.func(ri.instance, self)
            return self.state.fold()

    def semantics_call(self, ri: RuleInfo, node: Any, pos: int) -> Any:
        if ri.is_name:
            self.validate_is_not_keyword(node)

        action = self.find_semantic_action(ri.name)
        if action:
            parseinfo = self.make_parseinfo(node, pos)
            return boundcall(
                action,
                {},
                node,
                *ri.params,
                parseinfo=parseinfo,
                **ri.kwparams,
            )
        else:
            return node

    def validate_is_not_keyword(self, name: Any) -> None:
        name_str = str(name)
        if self.config.ignorecase:
            name_str = name_str.upper()
        if name_str in self.keywords:
            raise self.newexcept(f'"{name_str}" is a reserved word', KeywordError)

    def make_parseinfo(self, name: str, pos: int) -> ParseInfo | None:
        if not self.config.parseinfo:
            return None
        endpos = self.pos
        return ParseInfo(
            cursor=self.cursor,
            rule=name,
            pos=pos,
            endpos=endpos,
            line=self.cursor.lineat(pos),
            endline=self.cursor.lineat(endpos),
            alerts=self.state.alerts,
        )

    def set_parseinfo(self, node: Any, name: str, pos: int):
        parseinfo = self.make_parseinfo(name, pos)
        if parseinfo is None:
            return
        elif hasattr(node, 'set_parseinfo'):
            node.set_parseinfo(parseinfo)
        elif hasattr(node, 'parseinfo'):
            node.parseinfo = parseinfo

    def save_result(self, key: MemoKey, result: RuleResult) -> None:
        if islist(result.node):
            result = result._replace(node=closedlist(result.node))
        self._results[key] = result

    def set_left_recursion_guard(self, key: MemoKey) -> None:
        if not self.config.left_recursion:
            return
        ex = self.newexcept(key.ruleinfo.name, excls=FailedLeftRecursion)
        self.memoize(key, ex)

    def clear_recursion_errors(self) -> None:
        def filter_func(_key: MemoKey, value: Any) -> bool:
            return isinstance(value, FailedLeftRecursion)

        prune_dict(self._memos, filter_func)

    def constant(self, literal: Any, capture: bool = True) -> Any:
        self.next_token()
        self.tracer.trace_match(self, literal)

        if not isinstance(literal, str):
            self.state.append(literal)
            return literal
        literal = str(literal)  # for type linters

        context: dict[str, Any] = (
            safe_builtins()
            | getattr(self.semantics, 'safe_context', lambda: {})()
            | self.ast
            if isinstance(self.ast, AST)
            else {}
        )

        expression = Undefined
        result = literal
        while result != expression:
            expression = result
            if not isinstance(expression, str):
                break

            expression = trim(expression)
            with suppress(ValueError, SyntaxError):
                result = stdlib_ast.literal_eval(expression.strip())
                assert result is not Undefined
                continue

            try:
                fstr_expression = f'''f{expression!r}'''

                if is_eval_safe(fstr_expression, context):
                    result = safe_eval(fstr_expression, context)

                if result == expression and is_eval_safe(expression, context):
                    # NOTE: No f'{xyz}' evaluations occurred
                    result = safe_eval(expression, context)
            except Exception as e:
                raise FailedSemantics(
                    f'Error evaluating constant {literal!r}: {e}',
                ) from e

        if capture:
            self.state.append(result)
        return result

    _constant = constant



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import cast

from asciitree import LeftAligned
from rich import print
from rich.tree import Tree


type ASCIITree = dict[str, str | ASCIITree]

tree: ASCIITree = {
    'TatSu': {'parsers': {'base.py': {}, 'ebnf.py': {}}, 'codegen': {'python.py': {}}}
}


def print_ascii_tree(tree: ASCIITree) -> None:
    tr = LeftAligned()
    print(tr(tree))


def to_rich_tree(tree: ASCIITree, rich_tree: Tree | None = None) -> Tree:
    if rich_tree is None:
        # Assume the top-level dict has one key as the root
        root_name = next(iter(tree))
        rich_tree = Tree(root_name)
        to_rich_tree(cast(ASCIITree, tree[root_name]), rich_tree)
        return rich_tree

    for key, value in tree.items():
        if isinstance(value, dict):
            to_rich_tree(value, rich_tree.add(key))
        else:
            rich_tree.add(f"{key}: {value}" if value else key)
    return rich_tree


if __name__ == "__main__":
    print("[bold green]ASCII Tree Rendering:[/bold green]")
    print_ascii_tree(tree)

    print("\n[bold green]Rich Tree Rendering:[/bold green]")
    rich_tree = to_rich_tree(tree)
    print(rich_tree)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""
Editor-friendly run-length compression for JSONL text.

Replaces runs of 5+ spaces with ``~N~`` markers, guarded so that
any natural ``~`` in the input becomes ``~~`` and round-trips
correctly. A text editor, a ``sed`` script, or ``grep`` can all
operate on the compressed form.
"""

from __future__ import annotations

import re


_COMPRESS_RE = re.compile(r"     +")

_EXPAND_RE = re.compile(r"~~|~(\d+)~")


def _compress_dispatch(m: re.Match) -> str:
    return f"~{len(m.group(0))}~"


def compress(s: str) -> str:
    s = s.replace("~", "~~")
    return _COMPRESS_RE.sub(_compress_dispatch, s)


def _expand_dispatch(m: re.Match) -> str:
    if m.group(0) == "~~":
        return "~"
    return " " * int(m.group(1))


def expand(s: str) -> str:
    return _EXPAND_RE.sub(_expand_dispatch, s)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .packet import HasID, Packet, PacketLike, WithID
from .queue import PacketzQueue


__all__ = [
    "HasID",
    "Packet",
    "PacketLike",
    "PacketzQueue",
    "WithID",
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from ast import literal_eval
from collections.abc import Iterable, Sequence
from typing import Any

from .. import peg as g
from ..contexts import ParseContext
from ..contexts.infos import ParseInfo
from ..exceptions import FailedSemantics
from ..objectmodel.builder import ModelBuilderSemantics
from ..util import WARNING_print, eval_escapes, flatten, re, trim


class GrammarSemantics(ModelBuilderSemantics):
    def __init__(self, name: str | None = None, context: ParseContext | None = None):
        super().__init__(
            basetype=g.Model,
            constructors=g.model_classes(),  # ty:ignore[invalid-argument-type]
        )
        self.name = name
        self.context = context
        self.rulemap: dict[str, g.Rule] = {}

    def set_context(self, context: ParseContext):
        self.context = context

    @classmethod
    def _validate_literal(cls, ast: Any):
        try:
            literal_eval(repr(str(ast)))
        except SyntaxError as e:
            raise FailedSemantics('literal string error: ' + str(e)) from e

    @classmethod
    def _validate_pattern(cls, ast: Any):
        cls._validate_literal(ast)
        try:
            re.compile(str(ast))
        except (TypeError, re.error) as e:
            raise FailedSemantics(f'"{ast!r}"pattern error: {e!s}') from e

    def token(self, ast: str) -> g.Token:
        token = ast
        if not token:
            raise FailedSemantics('empty token')
        literal_eval(repr(token))
        return g.Token(ast=token)

    def pattern(self, ast: str) -> g.Pattern:
        pattern = ast
        self._validate_literal(pattern)
        return g.Pattern(ast=pattern)

    def regexes(self, ast: Iterable[str]) -> Iterable[str]:
        pattern = ''.join(ast)
        self._validate_pattern(pattern)
        return ast

    def regex(self, ast: str) -> str:
        pattern = ast
        self._validate_pattern(pattern)
        return pattern

    def meta(self, ast: str, *_args) -> g.Meta:
        match ast:
            case "name":
                return g.NameMeta()
            case "int":
                return g.IntMeta()
            case "uint":
                return g.UIntMeta()
            case "float":
                return g.FloatMeta()
            case "bool":
                return g.BoolMeta()
            case _:
                raise FailedSemantics(f'unknown meta: {ast}')

    def deprecated_regex(self, ast: str, _parseinfo: ParseInfo | None = None):
        return self.regex(ast)
        # import warnings
        # if parseinfo:
        #     pi = parseinfo
        #     msg = (
        #         f'Deprecated syntax "?/../?" for regular expressions'
        #         f' at {pi.cursor.input.source} line {pi.line + 1}'
        #         f'\n?/"{ast}"/?'
        #     )
        # else:
        #     msg = 'Deprecated syntax "?/../? for regular expressions"'
        # warnings.warn(
        #     message=msg,
        #     category=DeprecationWarning,
        #     stacklevel=12,
        # )
        # return ast

    def string(self, ast):
        value = ast
        return eval_escapes(value)

    def multiline_string(self, ast):
        value = ast
        value = trim(value.strip()).rstrip()
        return eval_escapes(value)

    def hex(self, ast):
        return int(ast, 16)

    def float(self, ast):
        return float(ast)

    def int(self, ast):
        return int(ast)

    def none(self, _ast):
        return None

    def boolean(self, ast):
        return str(ast).lower() not in {'false', 'no', 'fail', '0'}

    # JSON
    def null(self, _ast):
        return None

    # JSON
    def true(self, _ast):
        return True

    # JSON
    def false(self, _ast):
        return False

    # JSON
    def number(self, ast):
        if isinstance(ast, str):
            return literal_eval(ast)
        return ast

    def cut_deprecated(self, _ast):
        WARNING_print('The use of >> for cut is deprecated. Use the ~ symbol instead.')
        return g.Cut()

    def override_single_deprecated(self, ast):
        WARNING_print('The use of @ for override is deprecated. Use @: instead')
        return g.Override(ast)

    def sequence(self, ast):
        seq = ast
        assert isinstance(seq, Sequence), str(seq)
        if len(seq) == 1:
            return seq[0]
        return g.Sequence(ast=ast)

    def choice(self, ast):
        return g.Choice(ast=ast)

    def new_name(self, name):
        if name in self.rulemap:
            raise FailedSemantics(f'rule "{name!s}" already defined')
        return name

    def known_name(self, name) -> str:
        if name not in self.rulemap:
            raise FailedSemantics(f'rule "{name!s}" not yet defined')
        return name

    def rule(self, ast):
        decorators = ast.decorators or []
        name = ast.name
        base = ast.base
        params = ast.params
        kwparams = dict(ast.kwparams) if ast.kwparams else {}

        if 'override' not in decorators and name in self.rulemap:
            self.new_name(name)
        elif 'override' in decorators:
            self.known_name(name)

        if not base:
            rule = g.Rule(
                ast=ast,
                name=name,
                params=params,
                kwparams=kwparams,
                decorators=decorators,
            )
        else:
            self.known_name(base)
            baserule = self.rulemap[base]
            rule = g.BasedRule(
                ast=ast,
                name=name,
                baserule=baserule,
                params=params,
                kwparams=kwparams,
                decorators=decorators,
            )

        self.rulemap[name] = rule
        return rule

    def rule_include(self, ast):
        name = str(ast)
        self.known_name(name)

        return g.RuleInclude(ast=ast, name=name)

    def grammar(self, ast):
        directives = {d.name: d.value for d in flatten(ast.directives) if d}
        for value in directives.values():
            literal_eval(repr(value))
        keywords = tuple(flatten(ast.keywords)) or ()

        if directives.get('whitespace') in {'None', 'False'}:
            # NOTE: use '' because None will _not_ override defaults in configuration
            directives['whitespace'] = ''

        name = self.name or directives.get('grammar')
        grammar = g.Grammar(
            name,
            list(self.rulemap.values()),
            directives=directives,
            keywords=keywords,
        )
        return grammar



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Iterable
from enum import Enum, auto

from ..basic import Cut
from ..choice import Choice
from ..model import Box, Model, Rule
from ..syntax import Call, Sequence


__all__ = ['mark_left_recursion']


def _callable_rule_ids(exp: Model, rule_index: dict[str, int]) -> list[int]:
    if isinstance(exp, Call):
        target = rule_index.get(exp.name)
        return [target] if target is not None else []

    if isinstance(exp, Choice):
        result: list[int] = []
        for opt in exp.options:
            result.extend(_callable_rule_ids(opt, rule_index))
        return result

    if isinstance(exp, Sequence):
        result = []
        for item in exp.sequence:
            if isinstance(item, Cut):
                continue
            result.extend(_callable_rule_ids(item, rule_index))
            if not _is_nullable_safe(item):
                break
        return result

    if isinstance(exp, Box):
        return _callable_rule_ids(exp.exp, rule_index)

    return []


def _is_nullable_safe(exp: Model) -> bool:
    if isinstance(exp, Call):
        return False

    if isinstance(exp, Sequence):
        return all(_is_nullable_safe(item) for item in exp.sequence)

    if isinstance(exp, Choice):
        return any(_is_nullable_safe(opt) for opt in exp.options)

    return exp.is_nullable()


def mark_left_recursion(rules: Iterable[Rule]) -> list[Rule]:
    class State(Enum):
        FIRST = auto()
        CUTOFF = auto()
        VISITED = auto()

    rules = list(rules)
    rule_index = {rule.name: i for i, rule in enumerate(rules)}

    # Reset initial structural properties
    for rule in rules:
        rule.is_lrec = False
        rule.is_memo = not rule.no_memo

    # Map the rule expressions directly to target rule integers
    edges = [_callable_rule_ids(rule.exp, rule_index) for rule in rules]

    # Depth tracking parameters
    depth = 0
    depth_stack: list[int] = [-1]
    rule_depth: dict[int, int] = {}
    rule_state: list[State] = [State.FIRST] * len(rules)

    def dfs(rule_id: int) -> None:
        nonlocal depth

        if rule_state[rule_id] == State.FIRST:
            rule_state[rule_id] = State.CUTOFF
        else:
            return

        # Frame boundaries setup for an already marked recursive branch
        is_already_lrec = rules[rule_id].is_lrec
        if is_already_lrec:
            depth_stack.append(depth)

        rule_depth[rule_id] = depth
        depth += 1

        try:
            for child_id in edges[rule_id]:
                dfs(child_id)

                # Evaluate loop closures against dynamic stack depth boundaries
                if (
                    rule_state[child_id] == State.CUTOFF
                    and rule_depth[child_id] > depth_stack[-1]
                ):
                    # Flag the loop closing point as left-recursive
                    rules[child_id].is_lrec = True

                    # Surgical un-memoization of loop context parameters
                    for active_id in rule_depth:
                        rules[active_id].is_memo = False

        finally:
            # Unwind scope depth tracking parameters cleanly
            if is_already_lrec:
                depth_stack.pop()
            rule_depth.pop(rule_id, None)
            depth -= 1
            rule_state[rule_id] = State.VISITED

    # Execute the depth-framed sweep over every rule
    for i in range(len(rules)):
        dfs(i)

    return [rule for rule in rules if rule.is_lrec]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Callable
from pathlib import Path
from typing import Any

from .. import peg as g
from ..boot import TatSuParserGenerator
from ..boot.bootparser import GRAMMAR_MODEL, TatSuBootstrapParser
from ..config import ParserConfig
from ..exceptions import ParseException
from ..input import Text
from ..ngcodegen.grammar_gen import parsermodel_gen
from ..ngcodegen.ngmodel_gen import modelgen
from ..ngcodegen.ngparser_gen import pythongen
from ..objectmodel import Node
from ..objectmodel.builder import (
    BuilderConfig,
    Constructor,
    ModelBuilderSemantics,
    TypeContainer,
)
from ..util import hasha


__compiled_grammar_cache: dict[tuple[str | None, str, int], g.Grammar] = {}


def boot_grammar() -> g.Grammar:

    return GRAMMAR_MODEL


def compile(
    grammar: str | Text,
    name: str | None = None,
    *,
    config: ParserConfig | None = None,
    filename: str | None = None,
    basetype: type | None = None,
    semantics: Any = None,
    asmodel: bool = False,
    builderconfig: BuilderConfig | None = None,
    synthok: bool = True,
    typedefs: list[TypeContainer] | None = None,
    constructors: list[Constructor] | None = None,
    **settings: Any,
) -> g.Grammar:
    filename = filename or settings.pop('source', None)
    ParserConfig.new(
        config=config,
        semantics=semantics,
        name=name,
        source=filename,
        **settings,
    )
    if isinstance(semantics, type):
        raise TypeError(
            f'semantics must be an object instance or None, not class {semantics!r}',
        )
    cache = __compiled_grammar_cache

    key = (name, hasha(grammar), id(semantics))
    if key in cache:
        model = cache[key]
    else:
        gen = TatSuParserGenerator(name, **settings)
        model = cache[key] = gen.parse(grammar, **settings)

    asmodel = not semantics and (
        asmodel
        or isinstance(builderconfig, BuilderConfig)
        or basetype is not None
        or typedefs is not None
        or constructors is not None
    )
    if semantics is not None:
        model.semantics = semantics
    elif asmodel:
        # HACK: cheating, but necessary for bw-compatibility
        builderconfig = BuilderConfig.new(
            config=builderconfig,
            synthok=synthok,
            basetype=basetype,
            typedefs=typedefs,
            constructors=constructors,
        )
        model.semantics = ModelBuilderSemantics(config=builderconfig)

    model.initialize()
    return model


def compile_to_parser(
    grammar: str | Text,
    name: str | None = None,
    *,
    config: ParserConfig | None = None,
    filename: str | None = None,
    basetype: type | None = None,
    semantics: Any = None,
    builderconfig: BuilderConfig | None = None,
    synthok: bool = True,
    typedefs: list[TypeContainer] | None = None,
    constructors: list[Constructor] | None = None,
    **settings: Any,
) -> g.Grammar:
    return compile(
        grammar,
        name=name,
        config=config,
        filename=filename,
        basetype=basetype,
        semantics=semantics,
        asmodel=True,
        builderconfig=builderconfig,
        synthok=synthok,
        typedefs=typedefs,
        constructors=constructors,
        **settings,
    )


def parse(
    grammar: str,
    text: str,
    /,
    *,
    config: ParserConfig | None = None,
    start: str | None = None,
    name: str | None = None,
    filename: str | None = None,
    semantics: Any | None = None,
    asmodel: bool = False,
    builderconfig: BuilderConfig | None = None,
    basetype: type | None = None,
    synthok: bool = True,
    typedefs: list[TypeContainer] | None = None,
    constructors: list[Constructor] | None = None,
    **settings: Any,
) -> Any:
    filename = filename or settings.pop('source', None)
    config = ParserConfig.new(
        config=config,
        start=start,
        name=name,
        source=filename,
        semantics=semantics,
        **settings,
    )
    model = compile(grammar, config=config, asmodel=asmodel)
    config.semantics = semantics or model.semantics

    asmodel = not config.semantics and (
        asmodel
        or isinstance(builderconfig, BuilderConfig)
        or basetype is not None
        or typedefs is not None
        or constructors is not None
    )
    if asmodel:
        builderconfig = BuilderConfig.new(
            config=builderconfig,
            synthok=synthok,
            basetype=basetype,
            typedefs=typedefs,
            constructors=constructors,
        )
        config.semantics = ModelBuilderSemantics(config=builderconfig)
    return model.parse(text, start=start, semantics=semantics, config=config)


def to_python_sourcecode(
    grammar: str,
    /,
    *,
    name: str | None = None,
    filename: str | None = None,
    config: ParserConfig | None = None,
    **settings: Any,
) -> str:
    filename = filename or settings.pop('source', None)
    config = ParserConfig.new(config=config, name=name, source=filename, **settings)
    model = compile(grammar, config=config, name=name, source=filename)
    return pythongen(model)


def to_python_model(
    grammar: str,
    /,
    *,
    name: str | None = None,
    filename: str | None = None,
    basetype: type = Node,
    config: ParserConfig | None = None,
    **settings: Any,
) -> str:
    filename = filename or settings.pop('source', None)
    config = ParserConfig.new(config=config, name=name, source=filename, **settings)
    model = compile(grammar, name=name, source=filename, config=config)
    return modelgen(model, basetype=basetype)


def to_parsermodel_sourcecode(
    grammar: str,
    /,
    *,
    name: str | None = None,
    filename: str | None = None,
    config: ParserConfig | None = None,
    **settings: Any,
) -> str:
    filename = filename or settings.pop('source', None)
    model = compile(grammar, config=config, name=name, source=filename, **settings)
    return parsermodel_gen(model, name=name)


def load_grammar(path: str | Path) -> g.Grammar:
    """Load a Grammar from an .ebnf or .json file."""

    p = Path(path)
    source = p.read_text(encoding="utf-8")
    if p.suffix == ".json":
        return load_json_grammar(source)
    else:
        return compile(source)


def load_json_grammar(source: str) -> g.Grammar:
    return g.Grammar.loads(source)


def to_grammar_json(grammar: str) -> Any:
    parser = TatSuBootstrapParser()
    model = parser.parse(grammar)
    return model.asjson()


# for backwards compatibility. Use `compile()` instead
def genmodel(
    *,
    name: str | None = None,
    grammar: str | None = None,
    semantics: type | None = None,
    config: ParserConfig | None = None,
    **settings: Any,
) -> g.Grammar:
    if grammar is None:
        raise ParseException('grammar is None')

    return compile(grammar, name=name, semantics=semantics, config=config, **settings)


def gencode(
    *,
    name: str | None = None,
    grammar: str,
    trace: bool = False,
    filename: str | None = None,
    codegen: Callable = pythongen,
    config: ParserConfig | None = None,
    **settings: Any,
) -> str:
    filename = filename or settings.pop('source', None)
    model = compile(
        grammar,
        name=name,
        source=filename,
        trace=trace,
        config=config,
        **settings,
    )
    return codegen(model)



async def worker_logic(bar_id: int, progress: MultiProgress):
    # The user just focuses on their own work and calls .update()
    for i in range(1, 101):
        await asyncio.sleep(0.05)
        progress.update(bar_id, i)


async def main():
    # Concurrency management is fully absorbed by your library here:
    async with MultiProgress(total_bars=3) as progress:
        # The user just runs their own workers parallelly
        await asyncio.gather(
            worker_logic(0, progress),
            worker_logic(1, progress),
            worker_logic(2, progress),
        )


if __name__ == "__main__":
    asyncio.run(main())



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from pathlib import Path

from .cfg import CLIConfig
from .fmt import render_grammar
from .global_opt import add_global_options
from .grammar_opt import add_grammar_options
from .lib import Results, load_grammar


def add_grammar_cmd(subparsers):
    grammar_parser = subparsers.add_parser(
        "grammar",
        help="Grammar transformations",
        add_help=True,
    )
    add_global_options(grammar_parser)
    add_grammar_options(grammar_parser)
    grammar_parser.add_argument(
        "grammar", help="Path to the grammar source (.ebnf or .json)"
    )
    return grammar_parser


def grammar_cmd(cfg: CLIConfig) -> Results:
    """Handle the ``grammar`` subcommand."""
    if cfg.grammar is None:
        raise ValueError("expected a grammar path")
    path = cfg.grammar
    if not Path(path).is_file():
        raise ValueError(f"expected a grammar file, got {path}")
    grammar = load_grammar(path)

    payload = render_grammar(
        grammar,
        cfg,
        name=Path(path).stem,
    )
    return [(str(path), payload)]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import builtins
import datetime
import importlib
import keyword
import os
import os.path
from pathlib import Path
from typing import Any


def startscript() -> str:
    import __main__ as main

    if main.__package__:
        return main.__package__
    elif isinstance(getattr(main, '__file__', None), str):
        return Path(main.__file__).name
    else:
        return 'unknown'


def startpath() -> Path:
    return Path(startscript()).absolute().parent


def is_posix() -> bool:
    return os.name == 'posix'


def identity(*args: Any) -> Any:
    if len(args) == 1:
        return args[0]
    return args


def format_if(fmt, values):
    return fmt % values if values else ''


def timestamp():
    return '.'.join(
        ('%2.2d' % t for t in datetime.datetime.now(datetime.UTC).utctimetuple()[:-2]),
    )


try:
    # noinspection PyUnusedImports
    import psutil
except ImportError:

    def memory_use():
        return 0

else:

    def memory_use():  # pyright: ignore[reportRedeclaration]
        process = psutil.Process(os.getpid())
        return process.memory_info().rss


def try_read(filename: str | Path) -> str:
    path = Path(filename)
    for e in ['utf-8', 'utf-16', 'latin-1', 'cp1252', 'ascii']:
        try:
            return path.read_text(encoding=e)
        except UnicodeError:
            pass
    raise ValueError(f"cannot find the encoding for '{filename}'")


def pathlist_from_patterns(patterns, ignore=None, base='.', sizesort=False):
    ignore = ignore or ()
    base = Path(base or '.').expanduser()

    filenames = set()
    for pattern in patterns or []:
        path = base / pattern
        if path.is_file():
            filenames.add(path)
            continue

        if path.is_dir():
            path += '/*'

        parts = path.parts[1:] if path.is_absolute() else path.parts
        joined_pattern = str(Path().joinpath(*parts))
        filenames.update(
            (p for p in Path(path.root).glob(joined_pattern) if not p.is_dir()),
        )

    filenames = list(filenames)

    def excluded(path):
        if any(path.match(ex) for ex in ignore):
            return True

        def part_matches(part):
            return any(Path(part).match(ex) for ex in ignore or ())

        return any(part_matches(part) for part in path.parts)

    if ignore:
        filenames = [fn for fn in filenames if not excluded(fn)]
    if sizesort:
        filenames.sort(key=lambda f: f.stat().st_size)

    return filenames


def short_relative_path(path: str | Path, base: str | Path = '.') -> Path:
    path = Path(path)
    base = Path(base)
    common = Path(os.path.commonpath([base.resolve(), path.resolve()]))

    if common == path.root:
        return path
    elif common == Path.home():
        up = Path('~')
    elif common == base:
        up = Path()
    else:
        n = len(base.parts) - len(common.parts)
        up = Path('../' * n)

    rel = up / path.resolve().relative_to(common)
    if len(str(rel)) < len(str(path)):
        return rel
    else:
        return path


def fqn(obj: Any) -> str:
    # by Gemini (2026-01-30)
    """Helper to safely retrieve the fully qualified name of a callable."""

    module = getattr(obj, '__module__', None) or getattr(obj, '__package__', None)
    qualname = getattr(obj, '__qualname__', None) or getattr(obj, '__name__', None)

    if module and qualname and module != builtins.__name__:
        return f'{module}.{qualname}'
    return qualname or str(obj)


def fqntype(fqn: str, default_module: str = 'builtins') -> type:
    if '.' in fqn:
        module_name, class_name = fqn.rsplit('.', 1)
    else:
        module_name = default_module
        class_name = fqn

    module = importlib.import_module(module_name)
    return getattr(module, class_name)


def isreserved(name) -> bool:
    return (
        keyword.iskeyword(name)
        or keyword.issoftkeyword(name)
        or name in {'type', 'list', 'dict', 'set'}
    )


def typename(obj: Any) -> str:
    """Return the name of the type of the given object."""
    if isinstance(obj, type):
        return obj.__name__
    return type(obj).__name__



#!/usr/bin/env python3
# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations


if __name__ == "__main__":
    import sys

    url = sys.stdin.read().strip()
    out = f"""
.. image:: {url}
    :alt: New CLI tool demo
    :width: 800px
"""
    print(out)
    if not sys.stdout.isatty():
        print(out, file=sys.stderr)
    sys.exit()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import logging
import os
import sys
import warnings
from typing import Any

from .strtools import prints


def TATSUDEBUG() -> int:
    return 10 * int(os.environ.get('TATSUDEBUG', logging.ERROR // 10))


LEVEL: int = TATSUDEBUG()


def set_debugging(level: int | str | None) -> None:
    global LEVEL  # noqa: PLW0603
    LEVEL = translate_level(level)


def translate_level(level: int | str | None) -> int:
    if isinstance(level, str):
        return getattr(logging, level.upper())
    elif isinstance(level, int):
        if level < 10:
            return level * 10
        return level
    return logging.ERROR


def eprint(*args: Any, **kwargs: Any) -> None:
    file = kwargs.pop('file', sys.stderr)
    kwargs.pop("highlight", None)
    kwargs.pop("markup", None)
    print(*args, file=file, **kwargs)


def DEBUG_print(*args: Any, **kwargs: Any) -> None:
    if not (__debug__ or LEVEL <= logging.DEBUG):
        return
    eprint(*args, **kwargs)


def INFO_print(*args: Any, **kwargs: Any) -> None:
    if LEVEL <= logging.INFO:
        return
    eprint('ⓘ', *args, **kwargs)


def WARNING_print(*args: Any, **kwargs: Any) -> None:
    if LEVEL <= logging.WARNING:
        return
    eprint('⚠', *args, **kwargs)
    warnings.warn(prints(*args, **kwargs), stacklevel=2)


def ERROR_print(
    *args: Any,
    extype: type[Exception] = RuntimeError,
    **kwargs: Any,
) -> Exception:
    WARNING_print(*args, **kwargs)
    return extype(prints(*args, **kwargs))



#  Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
#  SPDX-License-Identifier: BSD-4-Clause
#

# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .base import (
    NIL,
    Box,
    Grammar,
    Leaf,
    Model,
    ModelContext,
    Rule,
    Synth,
    Void,
    model_classes,
)
from .basic import (
    EOF,
    EOF_SYM,
    EOL,
    EOL_SYM,
    Alert,
    Comment,
    Constant,
    Cut,
    Dot,
    EOLComment,
    Fail,
    Token,
)
from .choice import (
    Choice,
    Option,
)
from .closure import (
    Closure,
    EmptyClosure,
    Gather,
    Join,
    PositiveClosure,
    PositiveGather,
    PositiveJoin,
)
from .deprecated import LeftJoin, RightJoin
from .math import _ref, ref
from .meta import BoolMeta, FloatMeta, IntMeta, Meta, NameMeta, UIntMeta
from .named import Named, NamedList, Override, OverrideList
from .pattern import Pattern
from .rulelike import BasedRule, RuleInclude
from .semantics import GrammarSemantics
from .syntax import (
    Call,
    Group,
    Lookahead,
    NegativeLookahead,
    Optional,
    Sequence,
    SkipGroup,
    SkipTo,
)


__all__ = [
    'EOL',
    'EOF',
    'EOF_SYM',
    'EOLComment',
    'EOL_SYM',
    'Alert',
    'BasedRule',
    'BoolMeta',
    'Call',
    'Choice',
    'Option',
    'Closure',
    'Comment',
    'Constant',
    'Cut',
    'Box',
    'Synth',
    'Dot',
    'EmptyClosure',
    'Fail',
    'Gather',
    'Grammar',
    'Group',
    'Join',
    'LeftJoin',
    'Lookahead',
    'Leaf',
    'Model',
    'ModelContext',
    'Named',
    'NamedList',
    'NameMeta',
    'UIntMeta',
    'IntMeta',
    'FloatMeta',
    'Meta',
    'NegativeLookahead',
    'NIL',
    'Optional',
    'Override',
    'OverrideList',
    'Pattern',
    'PositiveClosure',
    'PositiveGather',
    'PositiveJoin',
    'RightJoin',
    'Rule',
    'RuleInclude',
    'Sequence',
    'SkipGroup',
    'SkipTo',
    'GrammarSemantics',
    'Token',
    'Void',
    'model_classes',
    '_ref',
    'ref',
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import json  # noqa: F401  # type: ignore
import pprint  # noqa: F401 # type: ignore
from typing import Any

import pytest

import tatsu
from tatsu import peg as g
from tatsu.contexts.ast import AST
from tatsu.objectmodel import Node
from tatsu.util import hasha, typename
from tatsu.util.strtools import trim


def test_node_kwargs() -> None:
    class Atom(Node):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

        arguments: Any = None
        symbol: Any = None

    atom = Atom(symbol='foo')
    assert type(atom).__name__ == 'Atom'
    assert atom.ast is None
    assert atom.symbol == 'foo'

    atom = Atom(symbol='foo', ast={})
    assert atom.ast == {}
    assert atom.symbol is not None
    assert atom.symbol == 'foo'

    atom = Atom(ast=42, symbol='foo')
    assert atom.ast == 42
    assert atom.symbol is not None
    assert atom.symbol == 'foo'

    atom = Atom(ast={'bar': 1}, symbol='foo')
    assert atom.ast == {'bar': 1}

    with pytest.raises(TypeError, match=r'Overriding method children'):
        atom = Atom(children=[])
        assert atom


def test_children():
    grammar = r"""
        @@grammar::Calc


        start
            =
            expression $
            ;


        expression
            =
            | add:addition
            | sub:subtraction
            | term:term
            ;


        addition::Add
            =
            left:term op:'+' ~ right:expression
            ;


        subtraction::Subtract
            =
            left:term op:'-' ~ right:expression
            ;


        term
            =
            | mul:multiplication
            | div:division
            | factor:factor
            ;


        multiplication::Multiply
            =
            left:factor op:'*' ~ right:term
            ;


        division::Divide
            =
            left:factor '/' ~ right:term
            ;


        factor
            =
            | subexpression
            | number
            ;


        subexpression
            =
            '(' ~ @:expression ')'
            ;


        number::int
            =
            /\d+/
            ;
    """

    parser = tatsu.compile_to_parser(grammar)
    assert parser
    model = parser.parse('3 + 5 * ( 10 - 20 )')
    assert isinstance(model, AST)
    assert model['add']
    assert model['add'].children()
    assert type(model['add'].children()[0]).__name__ == 'Multiply'


def test_model_repr():
    node = Node()
    assert repr(node) == 'Node()'

    node = Node('hello')
    assert repr(node) == "Node('hello')"

    node = Node('hello')
    assert repr(node) == "Node('hello')"

    node = Node(ast='hello', ctx='world')
    assert repr(node) == "Node('hello')"

    with pytest.raises(ValueError, match=r'world='):
        Node(ast='hello', world='world')

    token = g.Token(ast='hello')
    assert repr(token) == "Token('hello')"

    token = g.Token(token='hello')
    assert repr(token) == "Token('hello')"

    g.Token('hello')
    assert repr(token) == "Token('hello')"

    with pytest.raises(TypeError, match=r'unexpected keyword argument'):
        g.Token(x='x')

    with pytest.raises(TypeError, match=r"required.*'name'"):
        g.Named()

    named = g.Named(name='foo')
    assert repr(named) == f"Named(name='foo', exp={g.NIL.__name__}())"

    named = g.Named(name='foo', ast='bar')
    assert repr(named) == "Named(name='foo', exp='bar')"

    named = g.Named(ast={'name': 'foo', 'exp': 'bar'})
    assert repr(named) == "Named(name='foo', exp='bar')"

    named = g.Named(name='foo', exp='bar')
    assert repr(named) == "Named(name='foo', exp='bar')"


def test_calc_repr():
    calc_grammar = r"""
        @@grammar::CALC

        start: expression $

        expression:
            | addition
            | subtraction
            | term

        addition[Add]: left=expression '+' ~ right=term

        subtraction[Subtract]: left=expression '-' ~ right=term

        term:
            | multiplication
            | division
            | factor

        multiplication[Multiply]: left=term '*' ~ right=factor

        division[Divide]: left=term '/' ~ right=factor

        factor:
            | '(' ~ =expression ')'
            | number

        number[int]: /\d+/
    """

    calc_repr = r"""
    Grammar(
      name='CALC',
      directives={'grammar': 'CALC'},
      keywords=(),
      rules=(
        Rule(
          name='start',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Sequence([Call('expression'), EOF()])
        ),
        Rule(
          name='expression',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=False,
          is_lrec=True,
          exp=Choice(
            [Option(Call('addition')), Option(Call('subtraction')), Option(Call('term'))]
          )
        ),
        Rule(
          name='addition',
          params=['Add'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=False,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='left', exp=Call('expression')),
              Token('+'),
              Cut(),
              Named(name='right', exp=Call('term'))
            ]
          )
        ),
        Rule(
          name='subtraction',
          params=['Subtract'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=False,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='left', exp=Call('expression')),
              Token('-'),
              Cut(),
              Named(name='right', exp=Call('term'))
            ]
          )
        ),
        Rule(
          name='term',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=False,
          is_lrec=True,
          exp=Choice(
            [
              Option(Call('multiplication')),
              Option(Call('division')),
              Option(Call('factor'))
            ]
          )
        ),
        Rule(
          name='multiplication',
          params=['Multiply'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=False,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='left', exp=Call('term')),
              Token('*'),
              Cut(),
              Named(name='right', exp=Call('factor'))
            ]
          )
        ),
        Rule(
          name='division',
          params=['Divide'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=False,
          is_lrec=False,
          exp=Sequence(
            [
              Named(name='left', exp=Call('term')),
              Token('/'),
              Cut(),
              Named(name='right', exp=Call('factor'))
            ]
          )
        ),
        Rule(
          name='factor',
          params=(),
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Choice(
            [
              Option(Sequence([Token('('), Cut(), Override(Call('expression')), Token(')')])),
              Option(Call('number'))
            ]
          )
        ),
        Rule(
          name='number',
          params=['int'],
          kwparams={},
          decorators=[],
          is_name=False,
          is_tokn=False,
          no_memo=False,
          no_stak=False,
          is_memo=True,
          is_lrec=False,
          exp=Pattern('\\d+')
        )
      )
    )
    """

    model = tatsu.compile_to_parser(calc_grammar)
    modelrepr = trim(repr(model)).rstrip()

    # HACK FIXME
    # from pathlib import Path
    # simple = tatsu.compile(calc_grammar, asmodel=False)
    # assert simple.asjson() == []
    # Path('calctree.json').write_text(simple.asjsons())
    # Path('./scripts/calcmodel.pynofmt').write_text(modelrepr)
    # Path('asjsonmodel.json').write_text(model.asjsons())

    refrepr = trim(calc_repr).rstrip()
    # assert hasha(modelrepr) == hasha(refrepr)
    assert modelrepr == refrepr
    emodel = eval(modelrepr, vars(g), {})  # type: ignore # noqa: S307
    assert isinstance(emodel, g.Grammar)
    assert isinstance(emodel.rules[0], g.Rule)
    assert emodel.name == 'CALC'
    assert emodel.rules[0].name == 'start'
    assert emodel.rules[-1].name == 'number'
    assert isinstance(emodel.rules[-1].exp, g.Pattern)
    assert hasha(repr(emodel).strip()) == hasha(refrepr)

    exp = emodel.parse('3 * 2 + 1', asmodel=True)
    assert typename(exp) == 'Add'
    assert typename(exp.left) == 'Multiply'  # type: ignore
    assert typename(exp.right) == 'int'  # type: ignore

    addmodel = emodel.rulemap['addition']
    assert isinstance(addmodel, g.Model)
    assert addmodel.grammar is emodel

    addresult = addmodel.parse('2 + 2', asmodel=False)
    assert addresult == {'left': '2', 'right': '2'}

    exp = emodel.parse('3', start='number', asmodel=True)
    assert isinstance(exp, int)
    assert exp == 3

    exp = emodel.parse('3', asmodel=False)
    assert isinstance(exp, str)
    assert exp == '3'

    # assert pprint.pformat(emodel).rstrip() == refrepr

    # test rule access by name
    assert isinstance(model.rule.start, g.Rule)
    assert isinstance(model.rule.addition, g.Rule)
    assert isinstance(model.rule.multiplication, g.Rule)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from pathlib import Path

import pytest

from tatsu import api, peg, railroads
from tatsu.util import trim

from .fixtures import PATH_TATSU_GRAMMAR


@pytest.mark.skip("TODO")
def test_railroads():
    grammar = Path(PATH_TATSU_GRAMMAR).read_text()
    model = api.compile(grammar)
    print('RAILROADS')
    railroads.draw(model)

    tracks = railroads.tracks(model)
    # assert len(tracks) == 250

    track0 = "start ●─grammar─■"
    assert tracks[0] == track0
    trackm2 = "eof[EOF] ●─'$' ✂ ──■"
    assert tracks[-2] == trackm2
    assert not tracks[-1].rstrip()


def test_per_node():
    grammar = r"""
        start: options

        options: number | 'hello'

        number: /\d+/
    """
    model = api.compile(grammar)
    assert isinstance(model, peg.Grammar)
    optrule = model.rulemap['options']
    expected = """
         options ●───┬─number──┬──■
                     └─'hello'─┘
    """
    assert optrule.railroads() == trim(expected)
